#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Model for Language Modeling."""

import logging
from typing import Dict, Type, Union

import sentencepiece as _sentencepiece
import pytorch_lightning as pl
import torch.optim as optim
from torch import Tensor
from transformers import (
    AutoConfig,
    AutoModel,
    AutoModelForCausalLM,
    AutoModelForMaskedLM,
    AutoTokenizer,
    BartConfig,
    BartForConditionalGeneration,
    T5Config,
    T5ForConditionalGeneration,
    XLNetLMHeadModel,
)

# sentencepiece has to be loaded before lightning to avoid segfaults
_sentencepiece

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class LMModule(pl.LightningModule):
    """Pytorch lightning model for LM training."""

    def __init__(
        self,
        model_args: Dict[str, Union[float, int, str]],
    ) -> None:
        """Construct an LM lightning module.

        Args:
            model_args: model's arguments.
        """
        super().__init__()

        self.model_args = model_args

        self.model: AutoModel
        self.tokenizer: AutoTokenizer

        self.cache_dir = None
        if "cache_dir" in model_args:
            self.cache_dir = model_args["cache_dir"]

        self.init_model()

    def init_model(self) -> None:
        """Initialize an AutoModel."""

        if self.model_args["model_name_or_path"] is not None:
            self.model = AutoModel.from_pretrained(
                self.model_args["model_name_or_path"],
                cache_dir=self.cache_dir,
            )
        else:
            config = AutoConfig.from_pretrained(
                self.model_args["model_config_name"], cache_dir=self.cache_dir
            )

            self.model = AutoModel.from_config(config)

            logger.info("Training from scratch")

    def forward(self, x: Tensor) -> Tensor:  # type: ignore
        """Forward pass on Transformer model.

        Args:
            x: tensor of shape (batch_size, seq_length) containing the input_ids.
        Returns:
            logits of the model.
        """
        return self.model(x).logits  # type:ignore

    def configure_optimizers(
        self,
    ) -> Dict[str, object]:
        """Create and return the optimizer.

        Returns:
            output (dict of str: Any):
                - optimizer: the optimizer used to update the parameter.
                - ls_scheduler: the scheduler used to reduce the learning rate in every epoch.
                - monitor: the metric that the scheduler will track over the training.
        """

        if not isinstance(self.model_args["lr"], float):
            raise ValueError("Learning rate should be float")

        if not isinstance(self.model_args["lr_decay"], float):
            raise ValueError("Learning rate decay rate should be float")

        optimizer = optim.AdamW(self.parameters(), lr=self.model_args["lr"])

        scheduler = optim.lr_scheduler.StepLR(optimizer, 1, self.model_args["lr_decay"])

        output = {
            "optimizer": optimizer,
            "lr_scheduler": scheduler,
            "monitor": "val_loss",
        }
        return output

    def training_step(self, batch: Dict[str, Tensor], batch_idx: int) -> Tensor:  # type: ignore
        """
        Training step which encompasses the forward pass and the computation of the loss value.

        Args:
            batch: dictionary containing the input_ids and optionally the token_type_ids and the attention_type.
            batch_idx: index of the current batch, unused.
        Returns:
            loss computed on the batch.
        """
        loss = self.model(**batch).loss  # type:ignore
        self.log("train_loss", loss)
        return loss

    def validation_step(self, batch: Dict[str, Tensor], batch_idx: int) -> Tensor:  # type: ignore
        """
        Validation step which encompasses the forward pass and the computation of the loss value.

        Args:
            batch: dictionary containing the input_ids and optionally the token_type_ids and the attention_type.
            batch_idx: index of the current batch, unused.
        Returns:
            loss computed on the batch.
        """
        loss = self.model(**batch).loss  # type:ignore
        self.log("val_loss", loss)
        return loss


class MLMModule(LMModule):
    """Pytorch lightning model for MLM training."""

    def init_model(self) -> None:
        """Initialize a MLM model."""

        if self.model_args["model_name_or_path"] is not None:
            self.model = AutoModelForMaskedLM.from_pretrained(
                self.model_args["model_name_or_path"], cache_dir=self.cache_dir
            )
        else:
            config = AutoConfig.from_pretrained(
                self.model_args["model_config_name"], cache_dir=self.cache_dir
            )

            self.model = AutoModelForMaskedLM.from_config(config)

            logger.info("Training from scratch")

        self.tokenizer = AutoTokenizer.from_pretrained(
            self.model_args["tokenizer"], use_fast=False
        )

        self.model.resize_token_embeddings(len(self.tokenizer))  # type: ignore


class CGMModule(LMModule):
    """Pytorch lightning model for conditional generation training."""

    def init_model(self) -> None:
        """Initialize a model for conditional generation."""

        if self.model_args["model_name_or_path"] is not None:
            if "t5" in self.model_args["model_name_or_path"]:  # type:ignore
                self.model = T5ForConditionalGeneration.from_pretrained(
                    self.model_args["model_name_or_path"],  # type:ignore
                    cache_dir=self.cache_dir,
                )
            elif "bart" in self.model_args["model_name_or_path"]:  # type:ignore
                self.model = BartForConditionalGeneration.from_pretrained(
                    self.model_args["model_name_or_path"],  # type:ignore
                    cache_dir=self.cache_dir,
                )
            else:
                raise ValueError(
                    f"{self.model_args['model_name_or_path']} is not supported for conditional generation training."
                )
        else:
            if "t5" in self.model_args["model_config_name"]:
                config = T5Config.from_pretrained(
                    self.model_args["model_config_name"], cache_dir=self.cache_dir
                )

                self.model = T5ForConditionalGeneration.from_config(config)
            elif "bart" in self.model_args["model_config_name"]:
                config = BartConfig.from_pretrained(
                    self.model_args["model_config_name"], cache_dir=self.cache_dir
                )

                self.model = BartForConditionalGeneration.from_config(config)
            else:
                raise ValueError(
                    f"{self.model_args['model_name_or_path']} is not supported for conditional generation training."
                )

            logger.info("Training from scratch")

        self.tokenizer = AutoTokenizer.from_pretrained(
            self.model_args["tokenizer"], use_fast=False
        )

        self.model.resize_token_embeddings(len(self.tokenizer))  # type: ignore


class CLMModule(LMModule):
    """Pytorch lightning model for CLM training."""

    def init_model(self) -> None:
        """Initialize a CLM model."""

        if self.model_args["model_name_or_path"] is not None:
            self.model = AutoModelForCausalLM.from_pretrained(
                self.model_args["model_name_or_path"], cache_dir=self.cache_dir
            )
        else:
            config = AutoConfig.from_pretrained(
                self.model_args["model_config_name"], cache_dir=self.cache_dir
            )

            self.model = AutoModelForCausalLM.from_config(config)

            logger.info("Training from scratch")

        self.tokenizer = AutoTokenizer.from_pretrained(
            self.model_args["tokenizer"],
            sep_token="<|sep|>",
            pad_token="<|pad|>",
            use_fast=False,
        )

        self.model.resize_token_embeddings(len(self.tokenizer))  # type: ignore


class PLMModule(LMModule):
    """Pytorch lightning model for PLM training."""

    def init_model(self) -> None:
        """Initialize a PLM model."""

        if self.model_args["model_name_or_path"] is not None:
            self.model = XLNetLMHeadModel.from_pretrained(
                self.model_args["model_name_or_path"],  # type:ignore
                cache_dir=self.cache_dir,
            )
        else:
            config = AutoConfig.from_pretrained(
                self.model_args["model_config_name"], cache_dir=self.cache_dir
            )

            self.model = XLNetLMHeadModel.from_config(config)

            logger.info("Training from scratch")

        self.tokenizer = AutoTokenizer.from_pretrained(
            self.model_args["tokenizer"], use_fast=False
        )

        self.model.resize_token_embeddings(len(self.tokenizer))  # type: ignore


LM_MODULE_FACTORY: Dict[str, Type[LMModule]] = {
    "lm": LMModule,
    "mlm": MLMModule,
    "clm": CLMModule,
    "cgm": CGMModule,
    "plm": PLMModule,
}



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""GuacaMol baselines training utilities."""

from dataclasses import dataclass, field
from typing import Any, Dict

from ..core import TrainingPipeline, TrainingPipelineArguments


class GuacaMolBaselinesTrainingPipeline(TrainingPipeline):
    """GuacaMol Baselines training pipelines."""

    def train(  # type: ignore
        self,
        training_args: Dict[str, Any],
        model_args: Dict[str, Any],
        dataset_args: Dict[str, Any],
    ) -> None:
        """Generic training function for GuacaMol Baselines training.

        Args:
            training_args: training arguments passed to the configuration.
            model_args: model arguments passed to the configuration.
            dataset_args: dataset arguments passed to the configuration.

        Raises:
            NotImplementedError: the generic trainer does not implement the pipeline.
        """
        raise NotImplementedError


@dataclass
class GuacaMolDataArguments(TrainingPipelineArguments):
    """Arguments related to data loading."""

    __name__ = "dataset_args"

    train_smiles_filepath: str = field(
        metadata={"help": "Path of SMILES file for Training."}
    )
    test_smiles_filepath: str = field(
        metadata={"help": "Path of SMILES file for Validation."}
    )


@dataclass
class GuacaMolSavingArguments(TrainingPipelineArguments):
    """Saving arguments related to GuacaMol trainer."""

    __name__ = "saving_args"

    model_filepath: str = field(metadata={"help": "Path to the model file."})
    model_config_filepath: str = field(
        metadata={"help": "Path to the model config file."}
    )



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Argument parser unit tests."""

from dataclasses import dataclass, field
from typing import Union

from gt4sd.cli.argument_parser import ArgumentParser


@dataclass
class TestArguments:

    int_arg: int = field(default=0)

    float_arg: float = field(default=0.0)

    str_arg: str = field(default="test")

    bool_arg: bool = field(default=True)

    int_none_arg: Union[int, None] = field(default=None)

    float_none_arg: Union[float, None] = field(default=None)

    str_none_arg: Union[str, None] = field(default=None)

    bool_none_arg: Union[bool, None] = field(default=None)


def test_int_default():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses([])

    assert isinstance(args[0].int_arg, int)
    assert args[0].int_arg == 0


def test_float_default():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses([])

    assert isinstance(args[0].float_arg, float)
    assert args[0].float_arg == 0.0


def test_str_default():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses([])

    assert isinstance(args[0].str_arg, str)
    assert args[0].str_arg == "test"


def test_bool_default():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses([])

    assert isinstance(args[0].bool_arg, bool)
    assert args[0].bool_arg is True


def test_int_assigned():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses(["--int_arg", "1"])

    assert isinstance(args[0].int_arg, int)
    assert args[0].int_arg == 1


def test_float_assigned():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses(["--float_arg", "1.0"])

    assert isinstance(args[0].float_arg, float)
    assert args[0].float_arg == 1.0


def test_str_assigned():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses(["--str_arg", "my_test"])

    assert isinstance(args[0].str_arg, str)
    assert args[0].str_arg == "my_test"


def test_bool_assigned():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses(["--bool_arg", "False"])

    assert isinstance(args[0].bool_arg, bool)
    assert args[0].bool_arg is False


def test_bool_int_assigned():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses(["--bool_arg", "0"])

    assert isinstance(args[0].bool_arg, bool)
    assert args[0].bool_arg is False


def test_int_none():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses([])

    assert args[0].int_none_arg is None


def test_float_none():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses([])

    assert args[0].float_none_arg is None


def test_str_none():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses([])

    assert args[0].str_none_arg is None


def test_bool_none():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses([])

    assert args[0].bool_none_arg is None


def test_int_str_none():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses(["--int_none_arg", ""])

    assert args[0].int_none_arg is None


def test_float_str_none():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses(["--float_none_arg", ""])

    assert args[0].float_none_arg is None


def test_str_str_none():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses(["--str_none_arg", ""])

    assert args[0].str_none_arg is None


def test_bool_str_none():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses(["--bool_none_arg", ""])

    assert args[0].bool_none_arg is None



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import json
from typing import Any, Callable, Dict, List, Tuple, Type, Union

from rdkit import Chem
from modlamp.descriptors import GlobalDescriptor
from tdc import Oracle
from tdc.chem_utils.oracle.oracle import fp2fpfunc
from tdc.metadata import download_oracle_names

from ..domains.materials import MacroMolecule, SmallMolecule
from .core import PropertyValue
from .scores import SCORING_FACTORY


def to_mol(mol: SmallMolecule) -> Chem.Mol:
    """Safely convert a string or a rdkit.Chem.Mol to a rdkit.Chem.Mol.

    Args:
        mol: a string or a rdkit.Chem.Mol object.

    Raises:
        TypeError: if wrong type is given.

    Returns:
        a rdkit.Chem.Mol object.
    """
    if isinstance(mol, str):
        mol = Chem.MolFromSmiles(mol)
    elif isinstance(mol, Chem.Mol):
        pass
    else:
        raise TypeError(
            f"Please provide SMILES string or rdkit.Chem.Mol object not {type(mol)}"
        )
    return mol


def to_smiles(mol: SmallMolecule) -> str:
    """Safely returns a SMILES string from a SMILES or a rdkit.Chem.Mol object.

    Args:
        SmallMolecule: either a SMILES or a rdkit.Chem.Mol object.

    Returns:
        a SMILES string.
    """
    if isinstance(mol, str):
        try:
            mol = Chem.MolFromSmiles(mol)
        except Exception:
            raise ValueError(
                f"Could not convert SMILES string to rdkit.Chem.Mol: {mol}"
            )
    elif isinstance(mol, Chem.Mol):
        pass
    else:
        raise TypeError(
            f"Pass a SMILES string or rdkit.Chem.Mol object not {type(mol)}"
        )

    return Chem.MolToSmiles(mol, canonical=True)


def get_similarity_fn(
    target_mol: SmallMolecule, fp_key: str = "FCFP4"
) -> Callable[[SmallMolecule], PropertyValue]:
    """Get a similarity function for a target molecule.

    Args:
        target_mol: a target molecule as SMILES or rdkit.Chem.Mol object.
        fp_key: The type of fingerprint to use. One of `ECFP4`, `ECFP6`, `FCFP4` and `AP`.

    Returns:
        a similarity function that can be called with a `SmallMolecule`.
    """
    if fp_key not in fp2fpfunc.keys():
        raise ValueError(f"Choose fp_key from {fp2fpfunc.keys()}.")
    target_smiles = to_smiles(target_mol)
    return Oracle(name="similarity_meta", target_smiles=target_smiles, fp=fp_key)


def get_activity_fn(target: str) -> Callable[[SmallMolecule], PropertyValue]:
    """Get a function to measure activity/affinity against a protein target.

    Args:
        target: name of the target protein.

    Returns:
        an affinity function that can be called with a `SmallMolecule`.
    """
    if target not in download_oracle_names:
        raise ValueError(
            f"Supported targets are: {download_oracle_names}, not {target}"
        )
    return Oracle(name=target)


# for proteins
def get_sequence(protein: MacroMolecule) -> str:
    """Safely returns an amino acid sequence of a macromolecule

    Args:
        protein: either an AA sequence or a rdkit.Chem.Mol object that can be converted to FASTA.

    Raises:
        TypeError: if the input was none of the above types.
        ValueError: if the sequence was empty or could not be parsed into FASTA.

    Returns:
        an AA sequence.
    """
    if isinstance(protein, str):
        seq = protein.upper().strip()
        return seq
    elif isinstance(protein, Chem.Mol):
        seq = Chem.MolToFASTA(protein).split()
    else:
        raise TypeError(f"Pass a string or rdkit.Chem.Mol object not {type(protein)}")
    if seq == []:
        raise ValueError(
            f"Sequence was empty or rdkit.Chem.Mol could not be converted: {protein}"
        )
    return seq[-1]


def get_descriptor(protein: MacroMolecule) -> GlobalDescriptor:
    """Convert a macromolecule to a modlamp GlobalDescriptor object.

    Args:
        protein: either an AA sequence or a rdkit.Chem.Mol object that can be converted to FASTA.

    Returns:
        GlobalDescriptor object.
    """
    seq = get_sequence(protein)
    return GlobalDescriptor(seq)


def get_target_parameters(
    target: Union[str, Dict[str, Any]]
) -> Tuple[List[Type[Any]], List[float]]:
    """Generates a tuple of scorers and weight list

    Args:
        target: scoring functions and parameters related to it

    Return:
        A tuple containing scoring functions and weight list
    """
    score_list = []
    weights = []
    target_dictionary: Dict[str, Any] = {}
    if isinstance(target, str):
        target_dictionary = json.loads(target)
    elif isinstance(target, dict):
        target_dictionary = target
    else:
        raise ValueError(
            f"{target} of type {type(target)} is not supported: provide 'str' or 'Dict[str, Any]'"
        )
    for scoring_function_name, parameters in target_dictionary.items():
        weight = 1.0
        if "weight" in parameters:
            weight = parameters.pop("weight")
        score_list.append(SCORING_FACTORY[scoring_function_name](**parameters))
        weights.append(weight)
    return (score_list, weights)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Utilities used in the tests."""

from functools import lru_cache

from pydantic import BaseSettings


class GT4SDTestSettings(BaseSettings):
    """Utility variables for the tests setup."""

    gt4sd_s3_host: str = "s3.mil01.cloud-object-storage.appdomain.cloud"
    gt4sd_s3_access_key: str = "a19f93a1c67949f1a31db38e58bcb7e8"
    gt4sd_s3_secret_key: str = "5748375c761a4f09c30a68cd15e218e3b27ca3e2aebd7726"
    gt4sd_s3_secure: bool = True
    gt4sd_ci: bool = False

    class Config:
        # immutable and in turn hashable, that is required for lru_cache
        frozen = True

    @staticmethod
    @lru_cache(maxsize=None)
    def get_instance() -> "GT4SDTestSettings":
        return GT4SDTestSettings()



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from enum import Enum
from typing import List

from paccmann_generator.drug_evaluators import SIDER as _SIDER
from paccmann_generator.drug_evaluators import ClinTox as _ClinTox
from paccmann_generator.drug_evaluators import OrganDB as _OrganTox
from paccmann_generator.drug_evaluators import SCScore
from paccmann_generator.drug_evaluators import Tox21 as _Tox21
from pydantic import Field

from ...algorithms.core import (
    ConfigurablePropertyAlgorithmConfiguration,
    Predictor,
    PredictorAlgorithm,
)
from ...domains.materials import SmallMolecule
from ..core import (
    CallablePropertyPredictor,
    DomainSubmodule,
    PropertyPredictorParameters,
    PropertyValue,
    S3Parameters,
)
from ..utils import get_activity_fn, get_similarity_fn, to_smiles
from .functions import (
    bertz,
    esol,
    is_scaffold,
    lipinski,
    logp,
    molecular_weight,
    number_of_aromatic_rings,
    number_of_atoms,
    number_of_h_acceptors,
    number_of_h_donors,
    number_of_heterocycles,
    number_of_large_rings,
    number_of_rings,
    number_of_rotatable_bonds,
    number_of_stereocenters,
    plogp,
    qed,
    sas,
    tpsa,
)


# NOTE: property prediction parameters
class ScscoreConfiguration(PropertyPredictorParameters):
    score_scale: int = 5
    fp_len: int = 1024
    fp_rad: int = 2


class SimilaritySeedParameters(PropertyPredictorParameters):
    smiles: str = Field(..., example="c1ccccc1")
    fp_key: str = "ECFP4"


class ActivityAgainstTargetParameters(PropertyPredictorParameters):
    target: str = Field(..., example="drd2", description="name of the target.")


class S3ParametersMolecules(S3Parameters):
    domain: DomainSubmodule = DomainSubmodule("molecules")


class MCAParameters(S3ParametersMolecules):
    algorithm_name: str = "MCA"


class Tox21Parameters(MCAParameters):
    algorithm_application: str = "Tox21"


class ClinToxParameters(MCAParameters):
    algorithm_application: str = "ClinTox"


class SiderParameters(MCAParameters):
    algorithm_application: str = "SIDER"


class OrganToxParameters(MCAParameters):
    class Organs(str, Enum):
        adrenal_gland: str = "Adrenal Gland"
        bone_marrow: str = "Bone Marrow"
        brain: str = "Brain"
        eye: str = "Eye"
        heart: str = "Heart"
        kidney: str = "Kidney"
        liver: str = "Liver"
        lung: str = "Lung"
        lymph_node: str = "Lymph Node"
        mammary_gland: str = "Mammary Gland"
        ovary: str = "Ovary"
        pancreas: str = "Pancreas"
        pituitary_gland: str = "Pituitary Gland"
        spleen: str = "Spleen"
        stomach: str = "Stomach"
        testes: str = "Testes"
        thymus: str = "Thymus"
        thyroid_gland: str = "Thyroid Gland"
        urinary_bladder: str = "Urinary Bladder"
        uterus: str = "Uterus"

    class ToxType(str, Enum):
        chronic: str = "chronic"
        subchronic: str = "subchronic"
        multigenerational: str = "multigenerational"
        all: str = "all"

    algorithm_application: str = "OrganTox"
    site: Organs = Field(
        ...,
        example=Organs.kidney,
        description="name of the target site of interest.",
    )
    toxicity_type: ToxType = Field(
        default=ToxType.all,
        example=ToxType.chronic,
        description="type of toxicity for which predictions are made.",
        options=["chronic", "subchronic", "multigenerational", "all"],
    )


# NOTE: property prediction classes
class Plogp(CallablePropertyPredictor):
    """Calculate the penalized logP of a molecule. This is the logP minus the number of
    rings with > 6 atoms minus the SAS.
    """

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=plogp, parameters=parameters)


class Lipinski(CallablePropertyPredictor):
    """Calculate whether a molecule adheres to the Lipinski-rule-of-5.
    A crude approximation of druglikeness.
    """

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=lipinski, parameters=parameters)


class Esol(CallablePropertyPredictor):
    """Estimate the water solubility of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=esol, parameters=parameters)


class Scscore(CallablePropertyPredictor):
    """Calculate the synthetic complexity score (SCScore) of a molecule."""

    def __init__(
        self, parameters: ScscoreConfiguration = ScscoreConfiguration()
    ) -> None:
        super().__init__(
            callable_fn=SCScore(**parameters.dict()), parameters=parameters
        )


class Sas(CallablePropertyPredictor):
    """Calculate the synthetic accessibility score (SAS) for a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=sas, parameters=parameters)


class Bertz(CallablePropertyPredictor):
    """Calculate Bertz index of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=bertz, parameters=parameters)


class Tpsa(CallablePropertyPredictor):
    """Calculate the total polar surface area of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=tpsa, parameters=parameters)


class Logp(CallablePropertyPredictor):
    """Calculates the partition coefficient of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=logp, parameters=parameters)


class Qed(CallablePropertyPredictor):
    """Calculate the quantitative estimate of drug-likeness (QED) of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=qed, parameters=parameters)


class NumberHAcceptors(CallablePropertyPredictor):
    """Calculate number of H acceptors of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=number_of_h_acceptors, parameters=parameters)


class NumberAtoms(CallablePropertyPredictor):
    """Calculate number of atoms of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=number_of_atoms, parameters=parameters)


class NumberHDonors(CallablePropertyPredictor):
    """Calculate number of H donors of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=number_of_h_donors, parameters=parameters)


class NumberAromaticRings(CallablePropertyPredictor):
    """Calculate number of aromatic rings of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=number_of_aromatic_rings, parameters=parameters)


class NumberRings(CallablePropertyPredictor):
    """Calculate number of rings of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=number_of_rings, parameters=parameters)


class NumberRotatableBonds(CallablePropertyPredictor):
    """Calculate number of rotatable bonds of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=number_of_rotatable_bonds, parameters=parameters)


class NumberLargeRings(CallablePropertyPredictor):
    """Calculate the amount of large rings (> 6 atoms) of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=number_of_large_rings, parameters=parameters)


class MolecularWeight(CallablePropertyPredictor):
    """Calculate molecular weight of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=molecular_weight, parameters=parameters)


class IsScaffold(CallablePropertyPredictor):
    """Whether a molecule is identical to its Murcko Scaffold."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=is_scaffold, parameters=parameters)


class NumberHeterocycles(CallablePropertyPredictor):
    """The amount of heterocycles of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=number_of_heterocycles, parameters=parameters)


class NumberStereocenters(CallablePropertyPredictor):
    """The amount of stereo centers of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=number_of_stereocenters, parameters=parameters)


class SimilaritySeed(CallablePropertyPredictor):
    """Calculate the similarity of a molecule to a seed molecule."""

    def __init__(self, parameters: SimilaritySeedParameters) -> None:
        super().__init__(
            callable_fn=get_similarity_fn(
                target_mol=parameters.smiles, fp_key=parameters.fp_key
            ),
            parameters=parameters,
        )


class ActivityAgainstTarget(CallablePropertyPredictor):
    """Calculate the activity of a molecule against a target molecule."""

    def __init__(self, parameters: ActivityAgainstTargetParameters) -> None:
        super().__init__(
            callable_fn=get_activity_fn(target=parameters.target), parameters=parameters
        )


class _MCA(PredictorAlgorithm):
    """Base class for all MCA-based predictive algorithms."""

    def __init__(self, parameters: MCAParameters):

        # Set up the configuration from the parameters
        configuration = ConfigurablePropertyAlgorithmConfiguration(
            algorithm_type=parameters.algorithm_type,
            domain=parameters.domain,
            algorithm_name=parameters.algorithm_name,
            algorithm_application=parameters.algorithm_application,
            algorithm_version=parameters.algorithm_version,
        )

        # The parent constructor calls `self.get_model`.
        super().__init__(configuration=configuration)


class Tox21(_MCA):
    """Model to predict environmental toxicity for the 12 endpoints in Tox21."""

    def get_model(self, resources_path: str) -> Predictor:
        """Instantiate the actual model.

        Args:
            resources_path: local path to model files.

        Returns:
            Predictor: the model.
        """
        # This model returns a singular reward and not a prediction for all 12 classes.
        model = _Tox21(model_path=resources_path)

        # Wrapper to get toxicity-endpoint-level predictions
        def informative_model(x: SmallMolecule) -> List[PropertyValue]:
            x = to_smiles(x)
            _ = model(x)
            return model.predictions.detach().tolist()

        return informative_model

    @classmethod
    def get_description(cls) -> str:
        text = """
        This model predicts the 12 endpoints from the Tox21 challenge.
        The endpoints are: NR-AR, NR-AR-LBD, NR-AhR, NR-Aromatase, NR-ER, NR-ER-LBD, NR-PPAR-gamma, SR-ARE, SR-ATAD5, SR-HSE, SR-MMP, SR-p53
        For details on the data see: https://tripod.nih.gov/tox21/challenge/.
        """
        return text


class ClinTox(_MCA):
    """Model to predict environmental toxicity for the 12 endpoints in Tox21."""

    def get_model(self, resources_path: str) -> Predictor:
        """Instantiate the actual model.

        Args:
            resources_path: local path to model files.

        Returns:
            Predictor: the model.
        """
        # This model returns a singular reward and not a prediction for both classes.
        model = _ClinTox(model_path=resources_path)

        # Wrapper to get toxicity-endpoint-level predictions
        def informative_model(x: SmallMolecule) -> List[PropertyValue]:
            x = to_smiles(x)
            _ = model(x)
            return model.predictions.detach().tolist()

        return informative_model

    @classmethod
    def get_description(cls) -> str:
        text = """
        This model is a multitask classifier for two classes:
            1. Predicted probability to receive FDA approval.
            2. Predicted probability of failure in clinical trials.
        For details on the data see: https://pubs.rsc.org/en/content/articlehtml/2018/sc/c7sc02664a.
        """
        return text


class SIDER(_MCA):
    def get_model(self, resources_path: str) -> Predictor:
        """Instantiate the actual model.

        Args:
            resources_path: local path to model files.

        Returns:
            Predictor: the model.
        """
        # This model returns a singular reward and not a prediction for both classes.
        model = _SIDER(model_path=resources_path)

        # Wrapper to get toxicity-endpoint-level predictions
        def informative_model(x: SmallMolecule) -> List[PropertyValue]:
            x = to_smiles(x)
            _ = model(x)
            return model.predictions.detach().tolist()

        return informative_model

    @classmethod
    def get_description(cls) -> str:
        text = """
        This model is a multitask classifier to predict side effects of drugs across
        27 classes. For details on the data see:
        https://pubs.rsc.org/en/content/articlehtml/2018/sc/c7sc02664a.
        """
        return text


class OrganTox(_MCA):
    """Model to predict toxicity for different organs."""

    def __init__(self, parameters: OrganToxParameters) -> None:

        # Extract model-specific parameters
        self.site = parameters.site
        self.toxicity_type = parameters.toxicity_type

        super().__init__(parameters=parameters)

    def get_model(self, resources_path: str) -> Predictor:
        """Instantiate the actual model.

        Args:
            resources_path: local path to model files.

        Returns:
            Predictor: the model.
        """
        # This model returns a singular reward and not a prediction for both classes.
        model = _OrganTox(
            model_path=resources_path, site=self.site, toxicity_type=self.toxicity_type
        )

        # Wrapper to get toxicity-endpoint-level predictions
        def informative_model(x: SmallMolecule) -> List[PropertyValue]:
            x = to_smiles(x)
            _ = model(x)
            all_preds = model.predictions.detach()

            return all_preds[model.class_indices].tolist()

        return informative_model

    @classmethod
    def get_description(cls) -> str:
        text = """
        This model is a multitask classifier to toxicity across different organs and
        toxicity types (`chronic`, `subchronic`, `multigenerational` or `all`).
        The organ has to specified in the constructor. The toxicity type defaults to
        `all` in which case three values are returned in the order `chronic`,
        `multigenerational` and `subchronic`.

        For details on the data see:
        https://pubs.rsc.org/en/content/articlehtml/2018/sc/c7sc02664a.
        """
        return text



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Implementation details for huggingface diffusers generation algorithms.

Parts of the implementation inspired by: https://github.com/huggingface/diffusers/blob/main/examples/train_unconditional.py.
"""

import logging
import os
from typing import Any, List, Optional, Union

import numpy as np
import torch
from diffusers import (
    DDIMPipeline,
    DDIMScheduler,
    DDPMPipeline,
    DDPMScheduler,
    LDMPipeline,
    LDMTextToImagePipeline,
    LMSDiscreteScheduler,
    ScoreSdeVePipeline,
    ScoreSdeVeScheduler,
    StableDiffusionPipeline,
)

from ....frameworks.torch import device_claim

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def set_seed(seed: int = 42) -> None:
    """Set seed for all random number generators.

    Args:
        seed: seed to set. Defaults to 42.
    """
    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available:
        torch.cuda.manual_seed_all(seed)  # type:ignore


MODEL_TYPES = {
    "diffusion": DDPMPipeline,
    "diffusion_implicit": DDIMPipeline,
    "latent_diffusion": LDMPipeline,
    "latent_diffusion_conditional": LDMTextToImagePipeline,
    "stable_diffusion": StableDiffusionPipeline,
    "score_sde": ScoreSdeVePipeline,
}

SCHEDULER_TYPES = {
    "ddpm": DDPMScheduler,
    "ddim": DDIMScheduler,
    "discrete": LMSDiscreteScheduler,
    "continuous": ScoreSdeVeScheduler,
}


class Generator:
    """Implementation of a generator."""

    def __init__(
        self,
        resources_path: str,
        model_type: str,
        model_name: str,
        scheduler_type: str,
        auth_token: bool = True,
        prompt: Optional[str] = None,
        device: Optional[Union[torch.device, str]] = None,
    ):
        """A Diffusers generation algorithm.

        Args:
            resources_path: path to the cache.
            model_type: type of the model.
            model_name: name of the model weights/version.
            scheduler_type: type of the schedule.
            auth_token: authentication token for private models.
            prompt: target text to use for conditional generation.
            device: device where the inference
                is running either as a dedicated class or a string. If not provided is inferred.
        """
        self.device = device_claim(device)
        self.resources_path = resources_path
        self.model_type = model_type
        self.model_name = model_name
        self.scheduler_type = scheduler_type
        self.prompt = prompt
        self.auth_token = auth_token
        self.load_model()

    def load_model(self) -> None:
        """Load a pretrained diffusion generative model."""

        try:
            model_class = MODEL_TYPES[self.model_type]
        except KeyError:
            raise KeyError(f"model type: {self.model_type} not supported")

        if (
            os.path.exists(self.resources_path)
            and len(os.listdir(self.resources_path)) > 0
        ):
            model_name_or_path = self.resources_path
        else:
            model_name_or_path = self.model_name

        if self.model_type == "stable_diffusion":
            self.model = model_class.from_pretrained(
                model_name_or_path,
                use_auth_token=self.auth_token,
            )
        else:
            self.model = model_class.from_pretrained(model_name_or_path)
        self.model.to(self.device)

    def sample(self, number_samples: int = 1) -> List[Any]:
        """Sample images with optional conditioning.

        Args:
            number_samples: number of images to generate.
        Returns:
            generated samples.
        """
        # if prompt is provided, use it
        if self.prompt:
            item = self.model(batch_size=number_samples, prompt=self.prompt)
        else:
            item = self.model(batch_size=number_samples)
        item = item["sample"]
        return item



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""AdvancedManufacturing tests."""

import pickle
from typing import ClassVar, Type

import pytest

from gt4sd.algorithms.controlled_sampling.advanced_manufacturing import (
    AdvancedManufacturing,
    CatalystGenerator,
)
from gt4sd.algorithms.core import AlgorithmConfiguration
from gt4sd.algorithms.registry import ApplicationsRegistry


def get_classvar_type(class_var):
    """Extract type from ClassVar type annotation: `ClassVar[T]] -> T`."""
    return class_var.__args__[0]


@pytest.mark.parametrize(
    "config_class, algorithm_type, domain, algorithm_name",
    [
        (
            CatalystGenerator,
            "controlled_sampling",
            "materials",
            AdvancedManufacturing.__name__,
        )
    ],
)
def test_config_class(
    config_class: Type[AlgorithmConfiguration],
    algorithm_type: str,
    domain: str,
    algorithm_name: str,
):
    assert config_class.algorithm_type == algorithm_type
    assert config_class.domain == domain
    assert config_class.algorithm_name == algorithm_name

    for keyword, type_annotation in config_class.__annotations__.items():
        if keyword in ("algorithm_type", "domain", "algorithm_name"):
            assert type_annotation.__origin__ is ClassVar  # type: ignore
            assert str == get_classvar_type(type_annotation)


@pytest.mark.parametrize(
    "config_class",
    [
        (CatalystGenerator),
    ],
)
def test_config_instance(config_class: Type[AlgorithmConfiguration]):
    config = config_class()  # type:ignore
    assert config.algorithm_application == config_class.__name__


@pytest.mark.parametrize(
    "config_class",
    [
        (CatalystGenerator),
    ],
)
def test_available_versions(config_class: Type[AlgorithmConfiguration]):
    versions = config_class.list_versions()
    assert "v0" in versions


@pytest.mark.parametrize(
    "config, example_target, algorithm",
    [
        (
            CatalystGenerator,
            10.0,
            AdvancedManufacturing,
        ),
        (
            CatalystGenerator,
            "10.0",
            AdvancedManufacturing,
        ),
    ],
)
def test_generation_via_import(config, example_target, algorithm):
    advanced_manufacturing = algorithm(
        configuration=config(number_of_steps=10, number_of_points=10),
        target=example_target,
    )
    items = list(advanced_manufacturing.sample(5))
    assert len(items) == 5


@pytest.mark.parametrize(
    "algorithm_application, target",
    [
        (
            CatalystGenerator.__name__,
            10.0,
        ),
        (
            CatalystGenerator.__name__,
            "10.0",
        ),
    ],
)
def test_generation_via_registry(target, algorithm_application):
    advanced_manufacturing = ApplicationsRegistry.get_application_instance(
        target=target,
        algorithm_type="controlled_sampling",
        domain="materials",
        algorithm_name=AdvancedManufacturing.__name__,
        algorithm_application=algorithm_application,
        generated_length=100,
        number_of_steps=10,
        number_of_points=10,
    )
    items = list(advanced_manufacturing.sample(5))
    assert len(items) == 5


@pytest.mark.parametrize(
    "config_class",
    [
        (CatalystGenerator),
    ],
)
def test_configuration_pickable(config_class: Type[AlgorithmConfiguration]):
    # implementation
    obj = config_class(algorithm_version="test")

    # ---
    import inspect

    inspect.getmodule(config_class)
    # ---
    pickled_obj = pickle.dumps(obj)
    restored_obj = pickle.loads(pickled_obj)
    assert restored_obj.algorithm_version == "test"
    assert restored_obj == obj

    # registered
    Config = ApplicationsRegistry.get_application(
        algorithm_type="controlled_sampling",
        domain="materials",
        algorithm_name=AdvancedManufacturing.__name__,
        algorithm_application=config_class.__name__,
    ).configuration_class

    obj = Config(algorithm_version="test")
    pickled_obj = pickle.dumps(obj)
    restored_obj = pickle.loads(pickled_obj)

    assert restored_obj.algorithm_version == "test"
    assert restored_obj == obj



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""VaeMlp implementation."""

from argparse import ArgumentParser
from typing import Any, Dict, Tuple

from torch import nn

from ....arg_parser.utils import str2bool
from ..base_model import GranularEncoderDecoderModel
from ..loss import LOSS_FACTORY
from ..module import MlpDecoder, MlpEncoder
from ..utils import KLAnnealer


class VaeMlp(GranularEncoderDecoderModel):
    """VaeMlp - variational encoder using MLP with Gaussian prior and approximate posterior."""

    def __init__(
        self,
        name: str,
        position: int,
        data: Dict[str, str],
        input_size_enc: int = 256,
        hidden_size_enc: int = 256,
        n_layers_enc: int = 2,
        activation_enc: str = "linear",
        dropout_enc: float = 0.0,
        hidden_size_dec: int = 256,
        n_layers_dec: int = 2,
        activation_dec: str = "linear",
        dropout_dec: float = 0.0,
        output_size_dec: int = 256,
        latent_size: int = 196,
        loss_function: str = "mse",
        kl_low: float = 0.0,
        kl_high: float = 0.1,
        kl_n_epochs: int = 100,
        kl_start_epoch: int = 0,
        **kwargs,
    ) -> None:
        """Construct VaeMlp.

        Args:
            name: model name.
            position: position of the model.
            data: data name mappings.
            input_size_enc: encoder input size. Defaults to 256.
            hidden_size_enc: encoder hidden size. Defaults to 256.
            n_layers_enc: number of layers for the encoder. Defaults to 2.
            activation_enc: activation function for the encoder. Defaults to "linear".
            dropout_enc: encoder dropout rate. Defaults to 0.0.
            hidden_size_dec: decoder hidden size. Defaults to 256.
            n_layers_dec: number of layers for the decoder. Defaults to 2.
            activation_dec: activation function for the decoder. Defaults to "linear".
            dropout_dec: decoder dropout rate. Defaults to 0.0.
            output_size_dec: decoder output size. Defaults to 256.
            latent_size: size of the latent space. Defaults to 196.
            loss_function: loss function. Defaults to "mse".
            kl_low: low KL weight.
            kl_high: high KL weight.
            kl_n_epochs: KL number of epochs.
            kl_start_epoch: KL starting epoch.

        Raises:
            ValueError: in case the provided loss function is not supported.
        """
        super().__init__(name=name, data=data)
        self.position = position
        self.input_key = f"{name}_{data['input']}"
        self.target_key = f"{name}_{data['target']}"

        self.latent_size = latent_size
        self.hidden_size_enc = hidden_size_enc
        self.hidden_size_dec = hidden_size_dec
        self.input_size_enc = input_size_enc
        self.hidden_size_enc = hidden_size_enc
        self.n_layers_enc = n_layers_enc
        self.activation_enc = activation_enc
        self.dropout_enc = dropout_enc
        self.hidden_size_dec = hidden_size_dec
        self.n_layers_dec = n_layers_dec
        self.activation_dec = activation_dec
        self.dropout_dec = dropout_dec
        self.output_size_dec = output_size_dec

        self.loss_function_name = loss_function.lower()
        if self.loss_function_name not in LOSS_FACTORY:
            raise ValueError(
                f"loss_function={self.loss_function_name} not supported. Pick a valid one: {sorted(list(LOSS_FACTORY.keys()))}"
            )
        self.loss_function = LOSS_FACTORY[self.loss_function_name]

        self.encoder = MlpEncoder(
            input_size=input_size_enc,
            hidden_size=hidden_size_enc,
            output_size=hidden_size_enc,
            n_layers=n_layers_enc,
            activation=activation_enc,
            dropout=dropout_enc,
        )
        self.fc_mu = nn.Linear(hidden_size_enc, latent_size)
        self.fc_var = nn.Linear(hidden_size_enc, latent_size)
        self.decoder = MlpDecoder(
            latent_size=latent_size,
            hidden_size=hidden_size_dec,
            output_size=output_size_dec,
            n_layers=n_layers_dec,
            activation=activation_dec,
            dropout=dropout_dec,
        )

        self.epoch_counter = 0

        self.kl_annealer = KLAnnealer(
            kl_low=kl_low,
            kl_high=kl_high,
            n_epochs=kl_n_epochs,
            start_epoch=kl_start_epoch,
        )

    def decode(self, z: Any, *args, **kwargs) -> Any:
        """Decode a latent space point.

        Args:
            z: latent point.

        Returns:
            decoded sample.
        """
        return self.decoder(z)

    def _sampling_step(self, x: Any, *args, **kwargs) -> Any:
        """Run a sampling step in the model.

        Args:
            x: model input.

        Returns:
            model sampling step output.
        """
        x = self.encoder(x)
        mu = self.fc_mu(x)
        log_var = self.fc_var(x)
        return self.sample(mu, log_var)

    def encode(self, x: Any, *args, **kwargs) -> Any:
        """Encode a sample.

        Args:
            x: input sample.

        Returns:
            latent encoding.
        """
        _, _, z = self._sampling_step(x)
        return z

    def _run_step(self, x: Any, *args, **kwargs) -> Any:
        """Run a step in the model.

        Args:
            x: model input.

        Returns:
            model step output.
        """
        p, q, z = self._sampling_step(x)
        return z, self.decoder(z), p, q

    def step(
        self,
        input_data: Any,
        target_data: Any,
        device: str = "cpu",
        current_epoch: int = 0,
        *args,
        **kwargs,
    ) -> Tuple[Any, Any, Any]:
        """Training step for the model.

        Args:
            input_data: input for the step.
            target_data: target for the step.
            device: string representing the device to use. Defaults to "cpu".
            current_epoch: current epoch. Defaults to 0.

        Returns:
            a tuple containing the step output, the loss and the logs for the module.
        """
        x = input_data

        z, x_hat, p, q = self._run_step(x)

        x_hat = x_hat.view(-1, x_hat.size(-1))

        reconstruction_loss = self.loss_function(x_hat, x)

        log_qz = q.log_prob(z)
        log_pz = p.log_prob(z)
        kl_scaling_factor = self.kl_annealer(current_epoch)
        kl = log_qz - log_pz
        kl = kl.mean()
        kl_scaled = kl * kl_scaling_factor

        loss = kl_scaled + reconstruction_loss
        logs = {
            "reconstruction_loss": reconstruction_loss,
            "kl_scaled": kl_scaled,
            "kl_unscaled": kl,
            "kl_scaling_factor": kl_scaling_factor,
            "loss": loss,
        }

        return z, loss, logs

    def val_step(
        self,
        input_data: Any,
        target_data: Any,
        device: str = "cpu",
        current_epoch: int = 0,
        *args,
        **kwargs,
    ) -> Any:
        """Validation step for the model.

        Args:
            input_data: input for the step.
            target_data: target for the step.
            device: string representing the device to use. Defaults to "cpu".
            current_epoch: current epoch. Defaults to 0.

        Returns:
            a tuple containing the step output, the loss and the logs for the module.
        """
        return self.step(
            input_data=input_data,
            target_data=target_data,
            device=device,
            current_epoch=current_epoch,
        )

    @staticmethod
    def add_model_specific_args(
        parent_parser: ArgumentParser, name: str, *args, **kwargs
    ) -> ArgumentParser:
        """Adding to a parser model specific arguments.

        Args:
            parent_parser: patent parser.
            name: model name.

        Returns:
            updated parser.
        """
        parser = ArgumentParser(parents=[parent_parser], add_help=False)

        parser.add_argument(f"--data_path_{name}", type=str)
        parser.add_argument(f"--data_file_{name}", type=str)
        parser.add_argument(f"--dataset_type_{name}", type=str)
        parser.add_argument(f"--position_{name}", type=int, nargs="+")
        parser.add_argument(f"--input_{name}", type=str)
        parser.add_argument(f"--target_{name}", type=str)
        parser.add_argument(f"--checkpoint_path_{name}", type=str)
        parser.add_argument(f"--start_from_checkpoint_{name}", type=str2bool)
        parser.add_argument(f"--freeze_weights_{name}", type=str2bool)
        parser.add_argument(f"--checkpoint_model_name_{name}", type=str)
        parser.add_argument(f"--input_size_enc_{name}", type=int)
        parser.add_argument(f"--hidden_size_enc_{name}", type=int)
        parser.add_argument(f"--n_layers_enc_{name}", type=int)
        parser.add_argument(f"--activation_enc_{name}", type=str)
        parser.add_argument(f"--dropout_enc_{name}", type=float)
        parser.add_argument(f"--hidden_size_dec_{name}", type=int)
        parser.add_argument(f"--dropout_dec_{name}", type=float)
        parser.add_argument(f"--n_layers_dec_{name}", type=int)
        parser.add_argument(f"--activation_dec_{name}", type=str)
        parser.add_argument(f"--ouptput_size_enc_{name}", type=int)
        parser.add_argument(f"--loss_function_{name}", type=str)
        parser.add_argument(f"--latent_size_{name}", type=int)
        parser.add_argument(f"--kl_low_{name}", type=float)
        parser.add_argument(f"--kl_high_{name}", type=float)
        parser.add_argument(f"--kl_n_epochs_{name}", type=int)
        parser.add_argument(f"--kl_start_epoch_{name}", type=int)

        return parser



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import typing
from typing import List

import pkg_resources
import torch
from packaging import version
from torch.optim.lr_scheduler import (  # type: ignore
    ChainedScheduler,
    ConstantLR,
    CosineAnnealingLR,
    CosineAnnealingWarmRestarts,
    CyclicLR,
    ExponentialLR,
    LambdaLR,
    LinearLR,
    MultiplicativeLR,
    MultiStepLR,
    OneCycleLR,
    SequentialLR,
    StepLR,
    _LRScheduler,
)
from torch.utils.data.dataset import (
    ChainDataset,
    ConcatDataset,
    Dataset,
    IterableDataset,
    Subset,
    TensorDataset,
)

sane_datasets = [
    Dataset,
    ChainDataset,
    ConcatDataset,
    IterableDataset,
    Subset,
    TensorDataset,
]

torch_version = version.parse(pkg_resources.get_distribution("torch").version)

if torch_version < version.parse("1.12") and torch_version >= version.parse("1.10"):
    from torch.utils.data.dataset import DFIterDataPipe  # type: ignore

    sane_datasets.append(DFIterDataPipe)

if torch_version < version.parse("1.12") and torch_version >= version.parse("1.11"):
    from torch.utils.data.dataset import IterDataPipe, MapDataPipe  # type: ignore

    sane_datasets.extend([IterDataPipe, MapDataPipe])


sane_schedulers = [
    _LRScheduler,
    ChainedScheduler,
    ConstantLR,
    CosineAnnealingLR,
    CosineAnnealingWarmRestarts,
    CyclicLR,
    ExponentialLR,
    LambdaLR,
    LinearLR,
    MultiStepLR,
    MultiplicativeLR,
    OneCycleLR,
    SequentialLR,
    StepLR,
]


@typing.no_type_check
def fix_datasets(sane_datasets: List[Dataset]) -> None:
    """
    Helper function to revert TorchDrug dataset handling (which breaks core
    pytorch functionalities). For details see:
    https://github.com/DeepGraphLearning/torchdrug/issues/96

    Args:
        sane_datasets: A list of pytorch datasets.

    Raises:
        AttributeError: If a passed dataset was not sane.
    """
    dataset = sane_datasets[0]
    torch.utils.data.dataset.Dataset = dataset  # type: ignore
    torch.utils.data.dataset.ChainDataset = sane_datasets[1]  # type: ignore
    torch.utils.data.dataset.ConcatDataset = sane_datasets[2]  # type: ignore
    torch.utils.data.dataset.IterableDataset = sane_datasets[3]  # type: ignore
    torch.utils.data.dataset.Subset = sane_datasets[4]  # type: ignore
    torch.utils.data.dataset.TensorDataset = sane_datasets[5]  # type: ignore

    if torch_version < version.parse("1.12") and torch_version >= version.parse("1.10"):
        torch.utils.data.dataset.DFIterDataPipe = sane_datasets[6]  # type: ignore

    if torch_version < version.parse("1.12") and torch_version >= version.parse("1.11"):
        torch.utils.data.dataset.IterDataPipe = sane_datasets[7]  # type: ignore
        torch.utils.data.dataset.MapDataPipe = sane_datasets[8]  # type: ignore

    for ds in sane_datasets[1:]:
        if not issubclass(ds, dataset):
            raise AttributeError(
                f"Reverting silent TorchDrug overwriting failed, {ds} is not a subclass"
                f" of {dataset}."
            )


@typing.no_type_check
def fix_schedulers(sane_schedulers: List[_LRScheduler]) -> None:
    """
    Helper function to revert TorchDrug LR scheduler handling (which breaks core
    pytorch functionalities). For details see:
    https://github.com/DeepGraphLearning/torchdrug/issues/96

    Args:
        sane_schedulers: A list of pytorch lr_schedulers.

    Raises:
        AttributeError: If a passed lr_scheduler was not sane.
    """
    scheduler = sane_schedulers[0]
    torch.optim.lr_scheduler._LRScheduler = scheduler  # type: ignore
    torch.optim.lr_scheduler.ChainedScheduler = sane_schedulers[1]  # type: ignore
    torch.optim.lr_scheduler.ConstantLR = sane_schedulers[2]  # type: ignore
    torch.optim.lr_scheduler.CosineAnnealingLR = sane_schedulers[3]  # type: ignore
    torch.optim.lr_scheduler.CosineAnnealingWarmRestarts = sane_schedulers[4]  # type: ignore
    torch.optim.lr_scheduler.CyclicLR = sane_schedulers[5]  # type: ignore
    torch.optim.lr_scheduler.ExponentialLR = sane_schedulers[6]  # type: ignore
    torch.optim.lr_scheduler.LambdaLR = sane_schedulers[7]  # type: ignore
    torch.optim.lr_scheduler.LinearLR = sane_schedulers[8]  # type: ignore
    torch.optim.lr_scheduler.MultiStepLR = sane_schedulers[9]  # type: ignore
    torch.optim.lr_scheduler.MultiplicativeLR = sane_schedulers[10]  # type: ignore
    torch.optim.lr_scheduler.OneCycleLR = sane_schedulers[11]  # type: ignore
    torch.optim.lr_scheduler.SequentialLR = sane_schedulers[12]  # type: ignore
    torch.optim.lr_scheduler.StepLR = sane_schedulers[13]  # type: ignore

    for lrs in sane_schedulers[1:]:
        if not issubclass(lrs, scheduler):
            raise AttributeError(
                f"Reverting silent TorchDrug overwriting failed, {lrs} is not a subclass"
                f" of {scheduler}."
            )



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import inspect
import math
from collections import OrderedDict
from math import pi as PI
from operator import itemgetter
from typing import List, Optional, Set, Tuple

import numpy as np
import sympy as sym
import torch
import torch.nn as nn
from rdkit.Chem import AllChem
from scipy import special as sp
from scipy.optimize import brentq
from torch.nn import Linear, Sequential
from torch_geometric.data import Data
from torch_geometric.nn import global_add_pool, radius
from torch_geometric.utils import add_self_loops, remove_self_loops
from torch_scatter import scatter
from torch_sparse import SparseTensor

HAR2EV = 27.2113825435
KCALMOL2EV = 0.04336414


class MXMNetConfig(object):
    """MXMNet configuration."""

    def __init__(self, dim, n_layer, cutoff):
        """Initialize MXMNet configuration.

        Args:
            dim: dimension of the input.
            n_layer: number of layers.
            cutoff: cutoff radius.
        """
        self.dim = dim
        self.n_layer = n_layer
        self.cutoff = cutoff


class MXMNet(nn.Module):
    """MXMNet - Multiplex Molecular Graph Neural Network"""

    def __init__(
        self, config: MXMNetConfig, num_spherical=7, num_radial=6, envelope_exponent=5
    ) -> None:
        """Construct an MXMNet.

        Code adapted from: https://github.com/recursionpharma/gflownet/tree/trunk/src/gflownet/models and https://github.com/zetayue/MXMNet.

        Args:
            config: model configuration
            num_spherical: number of spherical harmonics to use.
            num_radial: number of radial harmonics to use.
            envelope_exponent: exponent of the envelope function.
        """
        super(MXMNet, self).__init__()

        self.name = "mxmnet"

        self.dim = config.dim
        self.n_layer = config.n_layer
        self.cutoff = config.cutoff

        self.embeddings = nn.Parameter(torch.ones((5, self.dim)))

        self.rbf_l = BesselBasisLayer(16, 5, envelope_exponent)
        self.rbf_g = BesselBasisLayer(16, self.cutoff, envelope_exponent)
        self.sbf = SphericalBasisLayer(num_spherical, num_radial, 5, envelope_exponent)

        self.rbf_g_mlp = MLP([16, self.dim])
        self.rbf_l_mlp = MLP([16, self.dim])

        self.sbf_1_mlp = MLP([num_spherical * num_radial, self.dim])
        self.sbf_2_mlp = MLP([num_spherical * num_radial, self.dim])

        self.global_layers = torch.nn.ModuleList()
        for layer in range(config.n_layer):
            self.global_layers.append(Global_MP(config))

        self.local_layers = torch.nn.ModuleList()
        for layer in range(config.n_layer):
            self.local_layers.append(Local_MP(config))

        self.init()

    def init(self):
        stdv = math.sqrt(3)
        self.embeddings.data.uniform_(-stdv, stdv)

    def indices(self, edge_index, num_nodes):
        """Compute indices.

        Args:
            edge_index: edge index of the graph.
            num_nodes: number of nodes in the graph.

        Returns:
            tuple of indeces.
        """
        row, col = edge_index

        value = torch.arange(row.size(0), device=row.device)
        adj_t = SparseTensor(
            row=col, col=row, value=value, sparse_sizes=(num_nodes, num_nodes)
        )

        # Compute the node indices for two-hop angles
        adj_t_row = adj_t[row]
        num_triplets = adj_t_row.set_value(None).sum(dim=1).to(torch.long)

        idx_i = col.repeat_interleave(num_triplets)
        idx_j = row.repeat_interleave(num_triplets)
        idx_k = adj_t_row.storage.col()
        mask = idx_i != idx_k
        idx_i_1, idx_j, idx_k = idx_i[mask], idx_j[mask], idx_k[mask]

        idx_kj = adj_t_row.storage.value()[mask]
        idx_ji_1 = adj_t_row.storage.row()[mask]

        # Compute the node indices for one-hop angles
        adj_t_col = adj_t[col]

        num_pairs = adj_t_col.set_value(None).sum(dim=1).to(torch.long)
        idx_i_2 = row.repeat_interleave(num_pairs)
        idx_j1 = col.repeat_interleave(num_pairs)
        idx_j2 = adj_t_col.storage.col()

        idx_ji_2 = adj_t_col.storage.row()
        idx_jj = adj_t_col.storage.value()

        return (
            idx_i_1,
            idx_j,
            idx_k,
            idx_kj,
            idx_ji_1,
            idx_i_2,
            idx_j1,
            idx_j2,
            idx_jj,
            idx_ji_2,
        )

    def forward(self, data):
        """Forward pass.

        Args:
            data: batch of data.

        Returns:
            gloabl pooled features.
        """
        x = data.x
        edge_index = data.edge_index
        pos = data.pos
        batch = data.batch
        # Initialize node embeddings
        h = torch.index_select(self.embeddings, 0, x.long())

        # Get the edges and pairwise distances in the local layer
        edge_index_l, _ = remove_self_loops(edge_index)
        j_l, i_l = edge_index_l
        dist_l = (pos[i_l] - pos[j_l]).pow(2).sum(dim=-1).sqrt()

        # Get the edges pairwise distances in the global layer
        row, col = radius(pos, pos, self.cutoff, batch, batch, max_num_neighbors=500)
        edge_index_g = torch.stack([row, col], dim=0)
        edge_index_g, _ = remove_self_loops(edge_index_g)
        j_g, i_g = edge_index_g
        dist_g = (pos[i_g] - pos[j_g]).pow(2).sum(dim=-1).sqrt()

        # Compute the node indices for defining the angles
        (
            idx_i_1,
            idx_j,
            idx_k,
            idx_kj,
            idx_ji,
            idx_i_2,
            idx_j1,
            idx_j2,
            idx_jj,
            idx_ji_2,
        ) = self.indices(edge_index_l, num_nodes=h.size(0))

        # Compute the two-hop angles
        pos_ji_1, pos_kj = pos[idx_j] - pos[idx_i_1], pos[idx_k] - pos[idx_j]
        a = (pos_ji_1 * pos_kj).sum(dim=-1)
        b = torch.cross(pos_ji_1, pos_kj).norm(dim=-1)
        angle_1 = torch.atan2(b, a)

        # Compute the one-hop angles
        pos_ji_2, pos_jj = pos[idx_j1] - pos[idx_i_2], pos[idx_j2] - pos[idx_j1]
        a = (pos_ji_2 * pos_jj).sum(dim=-1)
        b = torch.cross(pos_ji_2, pos_jj).norm(dim=-1)
        angle_2 = torch.atan2(b, a)

        # Get the RBF and SBF embeddings
        rbf_g = self.rbf_g(dist_g)
        rbf_l = self.rbf_l(dist_l)
        sbf_1 = self.sbf(dist_l, angle_1, idx_kj)
        sbf_2 = self.sbf(dist_l, angle_2, idx_jj)

        rbf_g = self.rbf_g_mlp(rbf_g)
        rbf_l = self.rbf_l_mlp(rbf_l)
        sbf_1 = self.sbf_1_mlp(sbf_1)
        sbf_2 = self.sbf_2_mlp(sbf_2)

        # Perform the message passing schemes
        node_sum = 0

        for layer in range(self.n_layer):
            h = self.global_layers[layer](h, rbf_g, edge_index_g)
            h, t = self.local_layers[layer](
                h, rbf_l, sbf_1, sbf_2, idx_kj, idx_ji, idx_jj, idx_ji_2, edge_index_l
            )
            node_sum += t

        # Readout
        output = global_add_pool(node_sum, batch)
        return output.view(-1)


class EMA:
    """EMA - Exponential Moving Average."""

    def __init__(self, model, decay):
        """Initialize ema.

        Args:
            model: model to ema.
            decay: decay rate.
        """
        self.decay = decay
        self.shadow = {}
        self.original = {}

        # Register model parameters
        for name, param in model.named_parameters():
            if param.requires_grad:
                self.shadow[name] = param.data.clone()

    def __call__(self, model, num_updates=99999):
        decay = min(self.decay, (1.0 + num_updates) / (10.0 + num_updates))
        for name, param in model.named_parameters():
            if param.requires_grad:
                assert name in self.shadow
                new_average = (1.0 - decay) * param.data + decay * self.shadow[name]
                self.shadow[name] = new_average.clone()

    def assign(self, model):
        for name, param in model.named_parameters():
            if param.requires_grad:
                assert name in self.shadow
                self.original[name] = param.data.clone()
                param.data = self.shadow[name]

    def resume(self, model):
        for name, param in model.named_parameters():
            if param.requires_grad:
                assert name in self.shadow
                param.data = self.original[name]


def MLP(channels) -> nn.Sequential:
    """multi-layer perceptron.

    Args:
        channels: list of number of channels.

    Returns:
        MLP model.
    """
    return Sequential(
        *[
            Sequential(Linear(channels[i - 1], channels[i]), SiLU())
            for i in range(1, len(channels))
        ]
    )


class Res(nn.Module):
    """Residual Block."""

    def __init__(self, dim):
        """Initialize residual block.

        Args:
            dim: dimension of the layer.
        """
        super(Res, self).__init__()

        self.mlp = MLP([dim, dim, dim])

    def forward(self, m):
        m1 = self.mlp(m)
        m_out = m1 + m
        return m_out


def compute_idx(pos, edge_index):
    """Compute the indices for the edges and angles.

    Args:
        pos: node positions.
        edge_index: edge indices.
    """

    pos_i = pos[edge_index[0]]
    pos_j = pos[edge_index[1]]

    d_ij = torch.norm(abs(pos_j - pos_i), dim=-1, keepdim=False).unsqueeze(-1) + 1e-5
    v_ji = (pos_i - pos_j) / d_ij

    unique, counts = torch.unique(
        edge_index[0], sorted=True, return_counts=True
    )  # Get central values
    full_index = (
        torch.arange(0, edge_index[0].size()[0]).cuda().int()
    )  # init full index

    # Compute 1
    repeat = torch.repeat_interleave(counts, counts)
    counts_repeat1 = torch.repeat_interleave(full_index, repeat)  # 0,...,0,1,...,1,...

    # Compute 2
    split = torch.split(full_index, counts.tolist())  # split full index
    index2 = list(edge_index[0].data.cpu().numpy())  # get repeat index
    counts_repeat2 = torch.cat(itemgetter(*index2)(split), dim=0)  # 0,1,2,...,0,1,2,..

    # Compute angle embeddings
    v1 = v_ji[counts_repeat1.long()]
    v2 = v_ji[counts_repeat2.long()]

    angle = (v1 * v2).sum(-1).unsqueeze(-1)
    angle = torch.clamp(angle, min=-1.0, max=1.0) + 1e-6 + 1.0

    return counts_repeat1.long(), counts_repeat2.long(), angle


def Jn(r, n):
    return np.sqrt(np.pi / (2 * r)) * sp.jv(n + 0.5, r)


def Jn_zeros(n, k):
    zerosj = np.zeros((n, k), dtype="float32")
    zerosj[0] = np.arange(1, k + 1) * np.pi
    points = np.arange(1, k + n) * np.pi
    racines = np.zeros(k + n - 1, dtype="float32")
    for i in range(1, n):
        for j in range(k + n - 1 - i):
            foo = brentq(Jn, points[j], points[j + 1], (i,))
            racines[j] = foo
        points = racines
        zerosj[i][:k] = racines[:k]

    return zerosj


def spherical_bessel_formulas(n):
    x = sym.symbols("x")
    f = [sym.sin(x) / x]
    a = sym.sin(x) / x
    for i in range(1, n):
        b = sym.diff(a, x) / x
        f += [sym.simplify(b * (-x) ** i)]
        a = sym.simplify(b)
    return f


def bessel_basis(n, k):
    zeros = Jn_zeros(n, k)
    normalizer = []

    for order in range(n):
        _normalizer_tmp = []
        for i in range(k):
            _normalizer_tmp += [0.5 * Jn(zeros[order, i], order + 1) ** 2]

        normalizer_tmp = 1 / np.array(_normalizer_tmp) ** 0.5
        normalizer += [normalizer_tmp]

    f = spherical_bessel_formulas(n)
    x = sym.symbols("x")
    bess_basis = []
    for order in range(n):
        bess_basis_tmp = []
        for i in range(k):
            bess_basis_tmp += [
                sym.simplify(
                    normalizer[order][i] * f[order].subs(x, zeros[order, i] * x)
                )
            ]
        bess_basis += [bess_basis_tmp]
    return bess_basis


def sph_harm_prefactor(k, m):
    return (
        (2 * k + 1)
        * math.factorial(k - abs(m))
        / (4 * np.pi * math.factorial(k + abs(m)))
    ) ** 0.5


def associated_legendre_polynomials(k, zero_m_only=True):
    z = sym.symbols("z")
    P_l_m = [[0] * (j + 1) for j in range(k)]

    P_l_m[0][0] = 1
    if k > 0:
        P_l_m[1][0] = z

        for j in range(2, k):
            P_l_m[j][0] = sym.simplify(
                ((2 * j - 1) * z * P_l_m[j - 1][0] - (j - 1) * P_l_m[j - 2][0]) / j
            )
        if not zero_m_only:
            for i in range(1, k):
                P_l_m[i][i] = sym.simplify((1 - 2 * i) * P_l_m[i - 1][i - 1])
                if i + 1 < k:
                    P_l_m[i + 1][i] = sym.simplify((2 * i + 1) * z * P_l_m[i][i])
                for j in range(i + 2, k):
                    P_l_m[j][i] = sym.simplify(
                        (
                            (2 * j - 1) * z * P_l_m[j - 1][i]
                            - (i + j - 1) * P_l_m[j - 2][i]
                        )
                        / (j - i)
                    )

    return P_l_m


def real_sph_harm(k, zero_m_only=True, spherical_coordinates=True):
    if not zero_m_only:
        S_m = [0]
        C_m = [1]
        for i in range(1, k):
            x = sym.symbols("x")
            y = sym.symbols("y")
            S_m += [x * S_m[i - 1] + y * C_m[i - 1]]
            C_m += [x * C_m[i - 1] - y * S_m[i - 1]]

    P_l_m = associated_legendre_polynomials(k, zero_m_only)
    if spherical_coordinates:
        theta = sym.symbols("theta")
        z = sym.symbols("z")
        for i in range(len(P_l_m)):
            for j in range(len(P_l_m[i])):
                if type(P_l_m[i][j]) != int:
                    P_l_m[i][j] = P_l_m[i][j].subs(z, sym.cos(theta))
        if not zero_m_only:
            phi = sym.symbols("phi")

            for i in range(len(S_m)):
                S_m[i] = (
                    S_m[i]  # type: ignore
                    .subs(x, sym.sin(theta) * sym.cos(phi))
                    .subs(y, sym.sin(theta) * sym.sin(phi))
                )
            for i in range(len(C_m)):
                C_m[i] = (
                    C_m[i]  # type: ignore
                    .subs(x, sym.sin(theta) * sym.cos(phi))
                    .subs(y, sym.sin(theta) * sym.sin(phi))
                )

    Y_func_l_m = [["0"] * (2 * j + 1) for j in range(k)]
    for i in range(k):
        Y_func_l_m[i][0] = sym.simplify(sph_harm_prefactor(i, 0) * P_l_m[i][0])

    if not zero_m_only:
        for i in range(1, k):
            for j in range(1, i + 1):
                Y_func_l_m[i][j] = sym.simplify(
                    2**0.5 * sph_harm_prefactor(i, j) * C_m[j] * P_l_m[i][j]
                )
        for i in range(1, k):
            for j in range(1, i + 1):
                Y_func_l_m[i][-j] = sym.simplify(
                    2**0.5 * sph_harm_prefactor(i, -j) * S_m[j] * P_l_m[i][j]
                )

    return Y_func_l_m


class BesselBasisLayer(torch.nn.Module):
    """Bessel Basis Layer."""

    def __init__(self, num_radial, cutoff, envelope_exponent=6) -> None:
        """Initialize Bessel basis layer.

        Args:
            num_radial: number of radial basis functions.
            cutoff: cutoff radius.
            envelope_exponent: envelope exponent.
        """

        super(BesselBasisLayer, self).__init__()
        self.cutoff = cutoff
        self.envelope = Envelope(envelope_exponent)

        self.freq = torch.nn.Parameter(torch.Tensor(num_radial))

        self.reset_parameters()

    def reset_parameters(self):
        torch.arange(1, self.freq.numel() + 1, out=self.freq.data).mul_(PI)

    def forward(self, dist):
        """Forward pass.

        Args:
            dist: distance matrix.

        Returns:
            Bessel basis.
        """
        dist = dist.unsqueeze(-1) / self.cutoff
        return self.envelope(dist) * (self.freq * dist).sin()


class SiLU(nn.Module):
    """SiLU Activation Function."""

    def __init__(self) -> None:
        """Initialize the SiLU activation function."""
        super().__init__()

    def forward(self, input):
        return silu(input)


def silu(input):
    return input * torch.sigmoid(input)


class Envelope(torch.nn.Module):
    """Envelope."""

    def __init__(self, exponent) -> None:
        """Initialize envelope.

        Args:
            exponent: exponent of the envelope.
        """
        super(Envelope, self).__init__()
        self.p = exponent
        self.a = -(self.p + 1) * (self.p + 2) / 2
        self.b = self.p * (self.p + 2)
        self.c = -self.p * (self.p + 1) / 2

    def forward(self, x):
        """Forward pass.

        Args:
            x: input.

        Returns:
            Envelope of x.
        """

        p, a, b, c = self.p, self.a, self.b, self.c
        x_pow_p0 = x.pow(p)
        x_pow_p1 = x_pow_p0 * x
        env_val = 1.0 / x + a * x_pow_p0 + b * x_pow_p1 + c * x_pow_p1 * x

        zero = torch.zeros_like(x)
        return torch.where(x < 1, env_val, zero)


class SphericalBasisLayer(torch.nn.Module):
    """Spherical Basis Layer."""

    def __init__(
        self, num_spherical, num_radial, cutoff=5.0, envelope_exponent=5
    ) -> None:
        """Initialize spherical basis layer.

        Args:
            num_spherical: number of spherical harmonics.
            num_radial: number of radial functions.
            cutoff: cutoff radius.
            envelope_exponent: envelope exponent.
        """
        super(SphericalBasisLayer, self).__init__()
        assert num_radial <= 64
        self.num_spherical = num_spherical
        self.num_radial = num_radial
        self.cutoff = cutoff
        self.envelope = Envelope(envelope_exponent)

        bessel_forms = bessel_basis(num_spherical, num_radial)
        sph_harm_forms = real_sph_harm(num_spherical)
        self.sph_funcs = []
        self.bessel_funcs = []

        x, theta = sym.symbols("x theta")
        modules = {"sin": torch.sin, "cos": torch.cos}
        for i in range(num_spherical):
            if i == 0:
                sph1 = sym.lambdify([theta], sph_harm_forms[i][0], modules)(0)
                self.sph_funcs.append(lambda x: torch.zeros_like(x) + sph1)
            else:
                sph = sym.lambdify([theta], sph_harm_forms[i][0], modules)
                self.sph_funcs.append(sph)
            for j in range(num_radial):
                bessel = sym.lambdify([x], bessel_forms[i][j], modules)
                self.bessel_funcs.append(bessel)

    def forward(self, dist, angle, idx_kj):
        """Forward pass.

        Args:
            dist: distance matrix.
            angle: angle matrix.
            idx_kj: index matrix.

        Returns:
            Spherical basis.
        """
        dist = dist / self.cutoff
        rbf = torch.stack([f(dist) for f in self.bessel_funcs], dim=1)
        rbf = self.envelope(dist).unsqueeze(-1) * rbf

        cbf = torch.stack([f(angle) for f in self.sph_funcs], dim=1)

        n, k = self.num_spherical, self.num_radial
        out = (rbf[idx_kj].view(-1, n, k) * cbf.view(-1, n, 1)).view(-1, n * k)
        return out


msg_special_args = set(
    ["edge_index", "edge_index_i", "edge_index_j", "size", "size_i", "size_j"]
)

aggr_special_args = set(["index", "dim_size"])

update_special_args: Set = set([])


class MessagePassing(torch.nn.Module):
    r"""Message Passing Layer.

    .. math::
        \mathbf{x}_i^{\prime} = \gamma_{\mathbf{\Theta}} \left( \mathbf{x}_i,
        \square_{j \in \mathcal{N}(i)} \, \phi_{\mathbf{\Theta}}
        \left(\mathbf{x}_i, \mathbf{x}_j,\mathbf{e}_{i,j}\right) \right),
    where :math:`\square` denotes a differentiable, permutation invariant
    function, *e.g.*, sum, mean or max, and :math:`\gamma_{\mathbf{\Theta}}`
    and :math:`\phi_{\mathbf{\Theta}}` denote differentiable functions such as
    MLPs.
    """

    def __init__(self, aggr="add", flow="target_to_source", node_dim=0) -> None:
        """Initialize message passing layer.

        Args:
            aggr: the aggregation scheme to use (add, mean, max).
            flow: the flow direction of message passing (source_to_target, target_to_source).
            node_dim: the axis along which to propagate.
        """
        super(MessagePassing, self).__init__()

        self.aggr = aggr
        assert self.aggr in ["add", "mean", "max"]

        self.flow = flow
        assert self.flow in ["source_to_target", "target_to_source"]

        self.node_dim = node_dim
        assert self.node_dim >= 0

        self.__msg_params__tmp = inspect.signature(self.message).parameters
        self.__msg_params__ = OrderedDict(self.__msg_params__tmp)

        self.__aggr_params__tmp = inspect.signature(self.aggregate).parameters
        self.__aggr_params__ = OrderedDict(self.__aggr_params__tmp)
        self.__aggr_params__.popitem(last=False)

        self.__update_params__tmp = inspect.signature(self.update).parameters
        self.__update_params__ = OrderedDict(self.__update_params__tmp)
        self.__update_params__.popitem(last=False)

        msg_args = set(self.__msg_params__.keys()) - msg_special_args
        aggr_args = set(self.__aggr_params__.keys()) - aggr_special_args
        update_args = set(self.__update_params__.keys()) - update_special_args

        self.__args__ = set().union(msg_args, aggr_args, update_args)

    def __set_size__(self, size, index, tensor):
        if not torch.is_tensor(tensor):
            pass
        elif size[index] is None:
            size[index] = tensor.size(self.node_dim)
        elif size[index] != tensor.size(self.node_dim):
            raise ValueError(
                (
                    f"Encountered node tensor with size {tensor.size(self.node_dim)} in dimension {self.node_dim}, but expected size {size[index]}."
                )
            )

    def __collect__(self, edge_index, size, kwargs):
        i, j = (0, 1) if self.flow == "target_to_source" else (1, 0)
        ij = {"_i": i, "_j": j}

        out = {}
        for arg in self.__args__:
            if arg[-2:] not in ij.keys():
                out[arg] = kwargs.get(arg, inspect.Parameter.empty)
            else:
                idx = ij[arg[-2:]]
                data = kwargs.get(arg[:-2], inspect.Parameter.empty)

                if data is inspect.Parameter.empty:
                    out[arg] = data
                    continue

                if isinstance(data, tuple) or isinstance(data, list):
                    assert len(data) == 2
                    self.__set_size__(size, 1 - idx, data[1 - idx])
                    data = data[idx]

                if not torch.is_tensor(data):
                    out[arg] = data
                    continue

                self.__set_size__(size, idx, data)
                out[arg] = data.index_select(self.node_dim, edge_index[idx])

        size[0] = size[1] if size[0] is None else size[0]
        size[1] = size[0] if size[1] is None else size[1]

        # Add special message arguments.
        out["edge_index"] = edge_index
        out["edge_index_i"] = edge_index[i]
        out["edge_index_j"] = edge_index[j]
        out["size"] = size
        out["size_i"] = size[i]
        out["size_j"] = size[j]

        # Add special aggregate arguments.
        out["index"] = out["edge_index_i"]
        out["dim_size"] = out["size_i"]

        return out

    def __distribute__(self, params, kwargs):
        out = {}
        for key, param in params.items():
            data = kwargs[key]
            if data is inspect.Parameter.empty:
                if param.default is inspect.Parameter.empty:
                    raise TypeError(f"Required parameter {key} is empty.")
                data = param.default
            out[key] = data
        return out

    def propagate(
        self, edge_index: torch.Tensor, size: Optional[List[Tuple]] = None, **kwargs
    ):
        """The initial call to start propagating messages.

        Args:
            edge_index: the indices of a general (sparse) assignment
                matrix with shape :obj:`[N, M]` (can be directed or
                undirected).
            size: the size :obj:`[N, M]` of the
                assignment matrix. If set to :obj:`None`, the size will be
                automatically inferred and assumed to be quadratic.
            **kwargs: any additional data which is needed to construct and
                aggregate messages, and to update node embeddings.
        """

        size = [None, None] if size is None else size  # type: ignore
        size = [size, size] if isinstance(size, int) else size  # type: ignore
        size = size.tolist() if torch.is_tensor(size) else size  # type: ignore
        size = list(size) if isinstance(size, tuple) else size
        assert isinstance(size, list)
        assert len(size) == 2

        kwargs = self.__collect__(edge_index, size, kwargs)

        msg_kwargs = self.__distribute__(self.__msg_params__, kwargs)

        m = self.message(**msg_kwargs)
        aggr_kwargs = self.__distribute__(self.__aggr_params__, kwargs)
        m = self.aggregate(m, **aggr_kwargs)

        update_kwargs = self.__distribute__(self.__update_params__, kwargs)
        m = self.update(m, **update_kwargs)

        return m

    def message(self, x_j):  # pragma: no cover
        r"""Constructs messages to node :math:`i` in analogy to
        :math:`\phi_{\mathbf{\Theta}}` for each edge in
        :math:`(j,i) \in \mathcal{E}` if :obj:`flow="source_to_target"` and
        :math:`(i,j) \in \mathcal{E}` if :obj:`flow="target_to_source"`.
        Can take any argument which was initially passed to :meth:`propagate`.
        In addition, tensors passed to :meth:`propagate` can be mapped to the
        respective nodes :math:`i` and :math:`j` by appending :obj:`_i` or
        :obj:`_j` to the variable name, *.e.g.* :obj:`x_i` and :obj:`x_j`.
        """

        return x_j

    def aggregate(self, inputs, index, dim_size):  # pragma: no cover
        r"""Aggregates messages from neighbors as
        :math:`\square_{j \in \mathcal{N}(i)}`.
        By default, delegates call to scatter functions that support
        "add", "mean" and "max" operations specified in :meth:`__init__` by
        the :obj:`aggr` argument.
        """

        return scatter(
            inputs, index, dim=self.node_dim, dim_size=dim_size, reduce=self.aggr
        )

    def update(self, inputs):  # pragma: no cover
        r"""Updates node embeddings in analogy to
        :math:`\gamma_{\mathbf{\Theta}}` for each node
        :math:`i \in \mathcal{V}`.
        Takes in the output of aggregation as first argument and any argument
        which was initially passed to :meth:`propagate`.
        """
        return inputs


params = AllChem.ETKDGv3()
params.useSmallRingTorsions = True


def rdkit_conformation(mol, n=5, addHs=False):
    """An function that finds the lowest energy conformation of a molecule.

    Args:
        mol: RDKit molecule object.
        n: Number of conformations to find.
        addHs: Whether to add hydrogens to the molecule.

    Returns:
        RDKit molecule object with lowest energy conformation. If none, no conformation is found.
    """
    if addHs:
        mol = AllChem.AddHs(mol)
    confs = AllChem.EmbedMultipleConfs(mol, numConfs=n, params=params)
    minc, aminc = 1000, 0
    for i in range(len(confs)):
        mp = AllChem.MMFFGetMoleculeProperties(mol, mmffVariant="MMFF94s")
        ff = AllChem.MMFFGetMoleculeForceField(mol, mp, confId=i)
        if ff is None:
            continue
        e = ff.CalcEnergy()
        if e < minc:
            minc = e
            aminc = i
    if len(confs):
        pos = []
        conf = mol.GetConformer(aminc)
        for i in range(mol.GetNumAtoms()):
            pos.append(list(conf.GetAtomPosition(i)))
        return torch.tensor(pos)
    return None


def mol2graph(mol):
    """Converts a RDKit molecule to a graph.

    Args:
        mol: RDKit molecule.

    Returns:
        A graph with node features and edge features.
    """
    mol = AllChem.AddHs(mol)
    N = mol.GetNumAtoms()
    try:
        pos = rdkit_conformation(mol)
        assert pos is not None, "no conformations found"
    except Exception:
        return None
    types = {"H": 0, "C": 1, "N": 2, "O": 3, "F": 4}
    type_idx = []
    for atom in mol.GetAtoms():
        type_idx.append(types[atom.GetSymbol()])

    row, col = [], []  # ,edge_type = [], [], []
    for bond in mol.GetBonds():
        start, end = bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()
        row += [start, end]
        col += [end, start]

    edge_index = torch.tensor([row, col], dtype=torch.long)
    perm = (edge_index[0] * N + edge_index[1]).argsort()
    edge_index = edge_index[:, perm]

    x = torch.tensor(type_idx).to(torch.float)
    data = Data(x=x, pos=pos, edge_index=edge_index)
    return data


class Global_MP(MessagePassing):
    """Global message passing."""

    def __init__(self, config) -> None:
        """Initializes the global message passing.

        Args:
            config: configuration.
        """
        super(Global_MP, self).__init__()
        self.dim = config.dim

        self.h_mlp = MLP([self.dim, self.dim])

        self.res1 = Res(self.dim)
        self.res2 = Res(self.dim)
        self.res3 = Res(self.dim)
        self.mlp = MLP([self.dim, self.dim])

        self.x_edge_mlp = MLP([self.dim * 3, self.dim])
        self.linear = nn.Linear(self.dim, self.dim, bias=False)

    def forward(self, h, edge_attr, edge_index):
        edge_index, _ = add_self_loops(edge_index, num_nodes=h.size(0))

        res_h = h

        # Integrate the Cross Layer Mapping inside the Global Message Passing
        h = self.h_mlp(h)

        # Message Passing operation
        h = self.propagate(edge_index, x=h, num_nodes=h.size(0), edge_attr=edge_attr)

        # Update function f_u
        h = self.res1(h)
        h = self.mlp(h) + res_h
        h = self.res2(h)
        h = self.res3(h)

        # Message Passing operation
        h = self.propagate(edge_index, x=h, num_nodes=h.size(0), edge_attr=edge_attr)

        return h

    def message(self, x_i, x_j, edge_attr, edge_index, num_nodes):
        num_edge = edge_attr.size()[0]

        x_edge = torch.cat((x_i[:num_edge], x_j[:num_edge], edge_attr), -1)
        x_edge = self.x_edge_mlp(x_edge)

        x_j = torch.cat((self.linear(edge_attr) * x_edge, x_j[num_edge:]), dim=0)

        return x_j

    def update(self, aggr_out):

        return aggr_out


class Local_MP(torch.nn.Module):
    """Local message passing."""

    def __init__(self, config) -> None:
        """Initialize local message passing.

        Args:
            config: configuration.
        """

        super(Local_MP, self).__init__()
        self.dim = config.dim

        self.h_mlp = MLP([self.dim, self.dim])

        self.mlp_kj = MLP([3 * self.dim, self.dim])
        self.mlp_ji_1 = MLP([3 * self.dim, self.dim])
        self.mlp_ji_2 = MLP([self.dim, self.dim])
        self.mlp_jj = MLP([self.dim, self.dim])

        self.mlp_sbf1 = MLP([self.dim, self.dim, self.dim])
        self.mlp_sbf2 = MLP([self.dim, self.dim, self.dim])
        self.lin_rbf1 = nn.Linear(self.dim, self.dim, bias=False)
        self.lin_rbf2 = nn.Linear(self.dim, self.dim, bias=False)

        self.res1 = Res(self.dim)
        self.res2 = Res(self.dim)
        self.res3 = Res(self.dim)

        self.lin_rbf_out = nn.Linear(self.dim, self.dim, bias=False)

        self.h_mlp = MLP([self.dim, self.dim])

        self.y_mlp = MLP([self.dim, self.dim, self.dim, self.dim])
        self.y_W = nn.Linear(self.dim, 1)

    def forward(
        self,
        h,
        rbf,
        sbf1,
        sbf2,
        idx_kj,
        idx_ji_1,
        idx_jj,
        idx_ji_2,
        edge_index,
        num_nodes=None,
    ):
        res_h = h

        # Integrate the Cross Layer Mapping inside the Local Message Passing
        h = self.h_mlp(h)

        # Message Passing 1
        j, i = edge_index
        m = torch.cat([h[i], h[j], rbf], dim=-1)

        m_kj = self.mlp_kj(m)
        m_kj = m_kj * self.lin_rbf1(rbf)
        m_kj = m_kj[idx_kj] * self.mlp_sbf1(sbf1)
        m_kj = scatter(m_kj, idx_ji_1, dim=0, dim_size=m.size(0), reduce="add")

        m_ji_1 = self.mlp_ji_1(m)

        m = m_ji_1 + m_kj

        # Message Passing 2       (index jj denotes j'i in the main paper)
        m_jj = self.mlp_jj(m)
        m_jj = m_jj * self.lin_rbf2(rbf)
        m_jj = m_jj[idx_jj] * self.mlp_sbf2(sbf2)
        m_jj = scatter(m_jj, idx_ji_2, dim=0, dim_size=m.size(0), reduce="add")

        m_ji_2 = self.mlp_ji_2(m)

        m = m_ji_2 + m_jj

        # Aggregation
        m = self.lin_rbf_out(rbf) * m
        h = scatter(m, i, dim=0, dim_size=h.size(0), reduce="add")

        # Update function f_u
        h = self.res1(h)
        h = self.h_mlp(h) + res_h
        h = self.res2(h)
        h = self.res3(h)

        # Output Module
        y = self.y_mlp(h)
        y = self.y_W(y)

        return h, y



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Custom exception definitions."""


class S3SyncError(RuntimeError):
    """Error in syncing the cache with S3."""

    def __init__(self, title: str, detail: str) -> None:
        """Initialize S3SyncError.

        Args:
            title: title of the error.
            detail: description of the error.
        """
        self.type = "S3SyncError"
        self.title = title
        self.detail = detail
        super().__init__(detail)


class InvalidItem(ValueError):
    """Error in validating an item."""

    def __init__(self, title: str, detail: str) -> None:
        """Initialize InvalidItem.

        Args:
            title: title of the error.
            detail: description of the error.
        """
        self.type = "InvalidItem"
        self.title = title
        self.detail = detail
        super().__init__(detail)


class InvalidAlgorithmConfiguration(ValueError):
    """Error in validating an algorithm configuration."""

    def __init__(self, title: str, detail: str) -> None:
        """Initialize InvalidAlgorithmConfiguration.

        Args:
            title: title of the error.
            detail: description of the error.
        """
        self.type = "InvalidAlgorithmConfiguration"
        self.title = title
        self.detail = detail
        super().__init__(detail)


class DuplicateApplicationRegistration(ValueError):
    """Error when identifier for a registration is not unique."""

    def __init__(self, title: str, detail: str) -> None:
        """Initialize DuplicateApplicationRegistration.

        Args:
            title: title of the error.
            detail: description of the error.
        """
        self.type = "InvalidAlgorithmConfiguration"
        self.title = title
        self.detail = detail
        super().__init__(detail)


class SamplingError(TimeoutError):
    """Error when inference takes too long."""

    def __init__(self, title: str, detail: str) -> None:
        """Initialize SamplingError.

        Args:
            title: title of the error.
            detail: description of the error.
        """
        self.type = "SamplingError"
        self.title = title
        self.detail = detail
        super().__init__(detail)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Module configuration."""

import logging
import os
from functools import lru_cache
from typing import Dict, Optional, Set

from pydantic import BaseSettings

from .s3 import GT4SDS3Client, S3SyncError, sync_folder_with_s3, upload_file_to_s3

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class GT4SDConfiguration(BaseSettings):
    """GT4SDConfiguration settings from environment variables.

    Default configurations for gt4sd including a read-only COS for algorithms' artifacts.
    Default configuration for gt4sd hub including a read-write COS for algorithms' artifacts uploaded by users.
    """

    gt4sd_local_cache_path: str = os.path.join(os.path.expanduser("~"), ".gt4sd")
    gt4sd_local_cache_path_algorithms: str = "algorithms"
    gt4sd_local_cache_path_properties: str = "properties"
    gt4sd_max_number_of_stuck_calls: int = 50
    gt4sd_max_number_of_samples: int = 1000000
    gt4sd_max_runtime: int = 86400
    gt4sd_create_unverified_ssl_context: bool = False

    gt4sd_s3_host: str = "s3.par01.cloud-object-storage.appdomain.cloud"
    gt4sd_s3_access_key: str = "6e9891531d724da89997575a65f4592e"
    gt4sd_s3_secret_key: str = "5997d63c4002cc04e13c03dc0c2db9dae751293dab106ac5"
    gt4sd_s3_secure: bool = True
    gt4sd_s3_bucket_algorithms: str = "gt4sd-cos-algorithms-artifacts"
    gt4sd_s3_bucket_properties: str = "gt4sd-cos-properties-artifacts"

    gt4sd_s3_host_hub: str = "s3.par01.cloud-object-storage.appdomain.cloud"
    gt4sd_s3_access_key_hub: str = "d9536662ebcf462f937efb9f58012830"
    gt4sd_s3_secret_key_hub: str = "934d1f3afdaea55ac586f6c2f729ac2ba2694bb8e975ee0b"
    gt4sd_s3_secure_hub: bool = True
    gt4sd_s3_bucket_hub_algorithms: str = "gt4sd-cos-hub-algorithms-artifacts"
    gt4sd_s3_bucket_hub_properties: str = "gt4sd-cos-hub-properties-artifacts"

    class Config:
        # immutable and in turn hashable, that is required for lru_cache
        frozen = True

    @staticmethod
    @lru_cache(maxsize=None)
    def get_instance() -> "GT4SDConfiguration":
        return GT4SDConfiguration()


class GT4SDArtifactManagementConfiguration:
    """Artifact management configuration."""

    gt4sd_s3_modules: Set[str] = {"algorithms", "properties"}

    def __init__(self, gt4sd_configuration: GT4SDConfiguration) -> None:
        """Initialize the artifact management configuration from the base one.

        Args:
            gt4sd_configuration: GT4SD base configuration.
        """
        self.local_cache_path: Dict[str, str] = {
            "algorithms": gt4sd_configuration.gt4sd_local_cache_path_algorithms,
            "properties": gt4sd_configuration.gt4sd_local_cache_path_properties,
        }
        self.s3_bucket: Dict[str, str] = {
            "algorithms": gt4sd_configuration.gt4sd_s3_bucket_algorithms,
            "properties": gt4sd_configuration.gt4sd_s3_bucket_properties,
        }
        self.s3_bucket_hub: Dict[str, str] = {
            "algorithms": gt4sd_configuration.gt4sd_s3_bucket_hub_algorithms,
            "properties": gt4sd_configuration.gt4sd_s3_bucket_hub_properties,
        }


gt4sd_configuration_instance = GT4SDConfiguration.get_instance()
gt4sd_artifact_management_configuration = GT4SDArtifactManagementConfiguration(
    gt4sd_configuration=gt4sd_configuration_instance
)

for key, val in gt4sd_artifact_management_configuration.local_cache_path.items():
    logger.info(f"using as local cache path for {key}: {val}")
    try:
        os.makedirs(val)
    except FileExistsError:
        logger.debug(f"local cache path for {key} already exists at {val}.")


def upload_to_s3(
    target_filepath: str, source_filepath: str, module: str = "algorithms"
):
    """Upload an algorithm in source_filepath in target_filepath on a bucket in the model hub.
    Args:
        target_filepath: path to save the objects in s3.
        source_filepath: path to the file to sync.
        module: the submodule of gt4sd that acts as a root for the bucket, defaults
            to `algorithms`.
    """

    if module not in gt4sd_artifact_management_configuration.gt4sd_s3_modules:
        raise ValueError(
            f"Unknown cache module: {module}. Supported modules: "
            f"{','.join(gt4sd_artifact_management_configuration.gt4sd_s3_modules)}"
        )

    try:
        upload_file_to_s3(
            host=gt4sd_configuration_instance.gt4sd_s3_host_hub,
            access_key=gt4sd_configuration_instance.gt4sd_s3_access_key_hub,
            secret_key=gt4sd_configuration_instance.gt4sd_s3_secret_key_hub,
            bucket=gt4sd_artifact_management_configuration.s3_bucket_hub[module],
            target_filepath=target_filepath,
            source_filepath=source_filepath,
            secure=gt4sd_configuration_instance.gt4sd_s3_secure_hub,
        )
    except S3SyncError:
        logger.exception("error in syncing the cache with S3")


def sync_algorithm_with_s3(
    prefix: Optional[str] = None, module: str = "algorithms"
) -> str:
    """Sync an algorithm in the local cache using environment variables.

    Args:
        prefix: the relative path in the bucket (both
            on S3 and locally) to match files to download. Defaults to None.
        module: the submodule of gt4sd that acts as a root for the bucket, defaults
            to `algorithms`.

    Returns:
        str: local path using the prefix.
    """
    if module not in gt4sd_artifact_management_configuration.gt4sd_s3_modules:
        raise ValueError(
            f"Unknown cache module: {module}. Supported modules: "
            f"{','.join(gt4sd_artifact_management_configuration.gt4sd_s3_modules)}"
        )

    folder_path = os.path.join(
        gt4sd_configuration_instance.gt4sd_local_cache_path,
        gt4sd_artifact_management_configuration.local_cache_path[module],
    )

    try:
        # sync with the public bucket
        sync_folder_with_s3(
            host=gt4sd_configuration_instance.gt4sd_s3_host,
            access_key=gt4sd_configuration_instance.gt4sd_s3_access_key,
            secret_key=gt4sd_configuration_instance.gt4sd_s3_secret_key,
            bucket=gt4sd_artifact_management_configuration.s3_bucket[module],
            folder_path=folder_path,
            prefix=prefix,
            secure=gt4sd_configuration_instance.gt4sd_s3_secure,
        )
        # sync with the public bucket hub
        sync_folder_with_s3(
            host=gt4sd_configuration_instance.gt4sd_s3_host_hub,
            access_key=gt4sd_configuration_instance.gt4sd_s3_access_key_hub,
            secret_key=gt4sd_configuration_instance.gt4sd_s3_secret_key_hub,
            bucket=gt4sd_artifact_management_configuration.s3_bucket_hub[module],
            folder_path=folder_path,
            prefix=prefix,
            secure=gt4sd_configuration_instance.gt4sd_s3_secure_hub,
        )
    except S3SyncError:
        logger.exception("error in syncing the cache with S3")
    return os.path.join(folder_path, prefix) if prefix is not None else folder_path


def get_cached_algorithm_path(
    prefix: Optional[str] = None, module: str = "algorithms"
) -> str:

    if module not in gt4sd_artifact_management_configuration.gt4sd_s3_modules:
        raise ValueError(
            f"Unknown cache module: {module}. Supported modules: "
            f"{','.join(gt4sd_artifact_management_configuration.gt4sd_s3_modules)}."
        )

    return (
        os.path.join(
            gt4sd_configuration_instance.gt4sd_local_cache_path,
            gt4sd_artifact_management_configuration.local_cache_path[module],
            prefix,
        )
        if prefix is not None
        else os.path.join(
            gt4sd_configuration_instance.gt4sd_local_cache_path,
            gt4sd_artifact_management_configuration.local_cache_path[module],
        )
    )


def get_algorithm_subdirectories_from_s3_coordinates(
    host: str,
    access_key: str,
    secret_key: str,
    bucket: str,
    secure: bool = True,
    prefix: Optional[str] = None,
) -> Set[str]:
    """Wrapper to initialize a client and list the directories in a bucket."""
    client = GT4SDS3Client(
        host=host, access_key=access_key, secret_key=secret_key, secure=secure
    )
    return client.list_directories(bucket=bucket, prefix=prefix)


def get_algorithm_subdirectories_with_s3(
    prefix: Optional[str] = None, module: str = "algorithms"
) -> Set[str]:
    """Get algorithms in the s3 buckets.

    Args:
        prefix: the relative path in the bucket (both
            on S3 and locally) to match files to download. Defaults to None.
        module: the submodule of gt4sd that acts as a root for the bucket, defaults
            to `algorithms`.

    Returns:
        Set: set of available algorithms on s3 with that prefix.
    """
    if module not in gt4sd_artifact_management_configuration.gt4sd_s3_modules:
        raise ValueError(
            f"Unknown cache module: {module}. Supported modules: "
            f"{','.join(gt4sd_artifact_management_configuration.gt4sd_s3_modules)}"
        )

    try:
        # directories in the read-only public bucket
        dirs = get_algorithm_subdirectories_from_s3_coordinates(
            host=gt4sd_configuration_instance.gt4sd_s3_host,
            access_key=gt4sd_configuration_instance.gt4sd_s3_access_key,
            secret_key=gt4sd_configuration_instance.gt4sd_s3_secret_key,
            bucket=gt4sd_artifact_management_configuration.s3_bucket[module],
            secure=gt4sd_configuration_instance.gt4sd_s3_secure,
            prefix=prefix,
        )

        # directories in the write public-hub bucket
        dirs_hub = get_algorithm_subdirectories_from_s3_coordinates(
            host=gt4sd_configuration_instance.gt4sd_s3_host_hub,
            access_key=gt4sd_configuration_instance.gt4sd_s3_access_key_hub,
            secret_key=gt4sd_configuration_instance.gt4sd_s3_secret_key_hub,
            bucket=gt4sd_artifact_management_configuration.s3_bucket_hub[module],
            secure=gt4sd_configuration_instance.gt4sd_s3_secure_hub,
            prefix=prefix,
        )

        # set of directories in the public bucket and public hub bucket
        versions = dirs.union(dirs_hub)
        return versions

    except Exception:
        logger.exception("generic syncing error")
        raise S3SyncError(
            "CacheSyncingError",
            f"error in getting directories of prefix={prefix}",
        )


def get_algorithm_subdirectories_in_cache(
    prefix: Optional[str] = None, module: str = "algorithms"
) -> Set[str]:
    """Get algorithm subdirectories from the cache.

    Args:
        prefix: prefix matching cache subdirectories. Defaults to None.
        module: the submodule of gt4sd that acts as a root for the bucket, defaults
            to `algorithms`.

    Returns:
        a set of subdirectories.
    """
    path = get_cached_algorithm_path(prefix=prefix, module=module)
    try:
        _, dirs, _ = next(iter(os.walk(path)))
        return set(dirs)
    except StopIteration:
        return set()


def reset_logging_root_logger():
    """Reset the root logger from logging library."""
    root = logging.getLogger()
    root.handlers = []
    root.filters = []



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Tests for registry that are independent of specific registrations."""

import pickle
from typing import ClassVar

import pytest
from pydantic import ValidationError

from gt4sd.algorithms.core import AlgorithmConfiguration, GeneratorAlgorithm
from gt4sd.algorithms.registry import ApplicationsRegistry
from gt4sd.exceptions import DuplicateApplicationRegistration

# there are at least 2 available versions, 1 per PaccMannRL configuration
AT_LEAST = 2


def assert_pickable(obj):
    pickled_obj = pickle.dumps(obj)
    restored_obj = pickle.loads(pickled_obj)

    assert restored_obj.algorithm_version == "test"
    assert restored_obj == obj

    return restored_obj


def test_list_available_s3():
    len(ApplicationsRegistry.list_available())
    assert len(ApplicationsRegistry.list_available()) >= AT_LEAST


def test_list_available_local_via_S3SyncError(mock_wrong_s3_env):
    assert len(ApplicationsRegistry.list_available()) >= AT_LEAST


def test_inherited_validation():
    Config = next(iter(ApplicationsRegistry.applications.values())).configuration_class
    with pytest.raises(
        ValidationError, match="algorithm_version\n +none is not an allowed value"
    ):
        Config(algorithm_version=None)  # type: ignore

    # NOTE: values convertible to string will not raise!
    Config(algorithm_version=5)  # type: ignore


def test_validation():
    with pytest.raises(
        ValidationError, match="batch_size\n +value is not a valid integer"
    ):
        ApplicationsRegistry.get_configuration_instance(
            algorithm_type="conditional_generation",
            domain="materials",
            algorithm_name="PaccMannRL",
            algorithm_application="PaccMannRLProteinBasedGenerator",
            batch_size="wrong_type",
        )


def test_pickable_wrapped_configurations():
    # https://github.com/samuelcolvin/pydantic/issues/2111
    Config = next(iter(ApplicationsRegistry.applications.values())).configuration_class
    restored_obj = assert_pickable(Config(algorithm_version="test"))

    # wrong type assignment, but we did not configure it to raise here:
    restored_obj.algorithm_version = object
    # ensure the restored dataclass is still a pydantic dataclass (mimic validation)
    _, optional_errors = restored_obj.__pydantic_model__.__fields__.get(
        "algorithm_version"
    ).validate(
        restored_obj.algorithm_version,
        restored_obj.__dict__,
        loc="algorithm_version",
        cls=restored_obj.__class__,
    )
    assert optional_errors is not None


def test_multiple_registration():
    class OtherAlgorithm(GeneratorAlgorithm):
        pass

    @ApplicationsRegistry.register_algorithm_application(
        GeneratorAlgorithm  # type:ignore
    )
    @ApplicationsRegistry.register_algorithm_application(OtherAlgorithm)  # type:ignore
    class Config(AlgorithmConfiguration):
        algorithm_type: ClassVar[str] = "dummy"
        domain: ClassVar[str] = ""
        algorithm_version: str = "development"

    # the class wrapping was applied twice
    config_class = ApplicationsRegistry.get_application(
        algorithm_type="dummy",
        domain="",
        algorithm_name="GeneratorAlgorithm",
        algorithm_application="Config",
    ).configuration_class
    assert config_class is Config
    assert config_class.algorithm_name == "GeneratorAlgorithm"
    assert config_class.algorithm_application == "Config"
    # __wrapped__?

    # retrieve singly wrapped config
    other_config_class = ApplicationsRegistry.get_application(
        algorithm_type="dummy",
        domain="",
        algorithm_name="OtherAlgorithm",
        algorithm_application="Config",
    ).configuration_class
    assert other_config_class is not Config
    assert other_config_class.algorithm_name == "OtherAlgorithm"
    assert other_config_class.algorithm_application == "Config"

    # registering Config directly and with explicit algorithm_application
    ExplicitConfig = ApplicationsRegistry.register_algorithm_application(
        GeneratorAlgorithm,  # type:ignore
        as_algorithm_application="ExplicitApplication",
    )(Config)
    explicit_config_class = ApplicationsRegistry.get_application(
        algorithm_type="dummy",
        domain="",
        algorithm_name="GeneratorAlgorithm",
        algorithm_application="ExplicitApplication",
    ).configuration_class
    assert explicit_config_class is ExplicitConfig
    assert explicit_config_class.algorithm_name == "GeneratorAlgorithm"
    assert explicit_config_class.algorithm_application == "ExplicitApplication"

    # overwriting value in applications is not allowed, applications are unique
    with pytest.raises(DuplicateApplicationRegistration):
        ApplicationsRegistry.register_algorithm_application(
            GeneratorAlgorithm  # type:ignore
        )(Config)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Enzeptional processing tests."""

from gt4sd.frameworks.enzeptional.processing import (
    reconstruct_sequence_with_mutation_range,
    sanitize_intervals,
)


def test_sanitize_intervals():
    assert sanitize_intervals([(-5, 12), (13, 14), (2, 3), (-3, 4), (-2, 6)]) == [
        (-5, 12),
        (13, 14),
    ]


def test_reconstruct_sequence_with_mutation_range():
    assert (
        reconstruct_sequence_with_mutation_range(
            "ABCDEFGHILMNOPQRSTUVWXYZ", "12789", [(0, 1), (6, 8)]
        )
        == "12CDEF789LMNOPQRSTUVWXYZ"
    )



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Tests for granular tokenizer."""

from gt4sd.frameworks.granular.tokenizer.tokenizer import (
    BigSmilesTokenizer,
    SelfiesTokenizer,
    SmilesTokenizer,
)


def test_tokenization():
    smiles = [
        "c1ccccc1",
        "c1ccc(CP(c2ccccc2)c2ccccc2)cc1.CCCCN1[C]N(Cc2ccccc2)c2ccccc21.[Ag]",
    ]

    def _test_tokenizer(tokenizer_type, tokens_groundtruth):
        tokenizer = tokenizer_type("test", smiles=smiles)
        tokens = tokenizer.tokenize(smiles[1])
        assert tokens_groundtruth == tokens
        assert [
            tokenizer.vocab[token] for token in tokens
        ] == tokenizer.convert_tokens_to_ids(tokenizer.tokenize(smiles[1]))
        assert 2 == len(
            [
                tokenizer.convert_tokens_to_ids(tokenizer.tokenize(a_smiles))
                for a_smiles in smiles
            ]
        )

    _test_tokenizer(
        SelfiesTokenizer,
        [
            "[c]",
            "[c]",
            "[c]",
            "[c]",
            "[Branch1_3]",
            "[=S]",
            "[C]",
            "[P]",
            "[Branch1_3]",
            "[Branch2_2]",
            "[c]",
            "[c]",
            "[c]",
            "[c]",
            "[c]",
            "[c]",
            "[Ring1]",
            "[Branch1_1]",
            "[c]",
            "[c]",
            "[c]",
            "[c]",
            "[c]",
            "[c]",
            "[Ring1]",
            "[Branch1_1]",
            "[c]",
            "[c]",
            "[Ring1]",
            "[#C]",
            "[.]",
            "[C]",
            "[C]",
            "[C]",
            "[C]",
            "[N]",
            "[Cexpl]",
            "[N]",
            "[Branch1_3]",
            "[Branch2_3]",
            "[C]",
            "[c]",
            "[c]",
            "[c]",
            "[c]",
            "[c]",
            "[c]",
            "[Ring1]",
            "[Branch1_1]",
            "[c]",
            "[c]",
            "[c]",
            "[c]",
            "[c]",
            "[c]",
            "[Ring1]",
            "[Branch1_1]",
            "[Ring1]",
            "[=N]",
            "[.]",
            "[Agexpl]",
        ],
    )

    _test_tokenizer(
        SmilesTokenizer,
        [
            "c",
            "1",
            "c",
            "c",
            "c",
            "(",
            "C",
            "P",
            "(",
            "c",
            "2",
            "c",
            "c",
            "c",
            "c",
            "c",
            "2",
            ")",
            "c",
            "2",
            "c",
            "c",
            "c",
            "c",
            "c",
            "2",
            ")",
            "c",
            "c",
            "1",
            ".",
            "C",
            "C",
            "C",
            "C",
            "N",
            "1",
            "[C]",
            "N",
            "(",
            "C",
            "c",
            "2",
            "c",
            "c",
            "c",
            "c",
            "c",
            "2",
            ")",
            "c",
            "2",
            "c",
            "c",
            "c",
            "c",
            "c",
            "2",
            "1",
            ".",
            "[Ag]",
        ],
    )


def test_big_smiles_tokenization():
    big_smiles = "{[][$]CC(C#N)[$],[$]CC(c1ccccc1)[$][]}"
    tokenizer = BigSmilesTokenizer("test")
    assert tokenizer.tokenize(big_smiles) == [
        "{",
        "[]",
        "[$]",
        "C",
        "C",
        "(",
        "C",
        "#",
        "N",
        ")",
        "[$]",
        ",",
        "[$]",
        "C",
        "C",
        "(",
        "c",
        "1",
        "c",
        "c",
        "c",
        "c",
        "c",
        "1",
        ")",
        "[$]",
        "[]",
        "}",
    ]



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Diffusion trainer unit tests."""

import os
import shutil
import tempfile
from typing import Any, Dict, cast

import pkg_resources

from gt4sd.training_pipelines import (
    TRAINING_PIPELINE_MAPPING,
    DiffusionForVisionTrainingPipeline,
)

TEST_DATA_DIRECTORY = pkg_resources.resource_filename(
    "gt4sd",
    "training_pipelines/tests/",
)


def _create_training_output_filepaths(directory: str) -> Dict[str, str]:
    """Create output filepath from directory.

    Args:
        directory: output directory.

    Returns:
        a dictionary containing the output files.
    """
    return {
        "config_save": os.path.join(directory, "config.pt"),
    }


template_config = {
    "model_args": {
        "model_path": "",
        "training_name": "ddpm",
        "num_train_timesteps": 1000,
        "learning_rate": 1e-4,
        "lr_scheduler": "cosine",
        "lr_warmup_steps": 500,
        "adam_beta1": 0.95,
        "adam_beta2": 0.999,
        "adam_weight_decay": 1e-6,
        "adam_epsilon": 1e-8,
        "gradient_accumulation_steps": 1,
        "in_channels": 3,
        "out_channels": 3,
        "layers_per_block": 2,
    },
    "training_args": {
        "local_rank": -1,
        "output_dir": "./outputs/",
        "save_images_epochs": 10,
        "save_model_epochs": 10,
        "mixed_precision": "no",
        "logging_dir": "./logs/",
        "cache_dir": "./cifar10/",
        "use_auth_token": False,
        "ema_inv_gamma": 1.0,
        "ema_power": 0.75,
        "ema_max_decay": 0.9999,
        "use_ema": True,
        "dummy_training": True,
        "is_sampling": False,
    },
    "dataset_args": {
        "dataset_name": "cifar10",
        "dataset_config_name": None,
        "resolution": 32,
        "train_batch_size": 16,
        "eval_batch_size": 16,
        "num_epochs": 1,
    },
}


def test_train():

    pipeline = TRAINING_PIPELINE_MAPPING.get("diffusion-trainer")

    assert pipeline is not None

    TEMPORARY_DIRECTORY = tempfile.mkdtemp()

    test_pipeline = cast(DiffusionForVisionTrainingPipeline, pipeline())

    config: Dict[str, Any] = template_config.copy()
    for key, value in _create_training_output_filepaths(TEMPORARY_DIRECTORY).items():
        config["training_args"][key] = value

    test_pipeline.train(**config)

    shutil.rmtree(TEMPORARY_DIRECTORY)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""NoEncoding implementation."""

from argparse import ArgumentParser
from typing import Any, Dict, Tuple

from ....arg_parser.utils import str2bool
from ..base_model import GranularEncoderDecoderModel


class NoEncoding(GranularEncoderDecoderModel):
    """NoEncoding module for adding inputs directly in the latent space."""

    def __init__(
        self,
        name: str,
        position: int,
        data: Dict[str, str],
        latent_size: int = 2,
        **kwargs,
    ) -> None:
        """Construct NoEncoding.

        Args:
            name: model name.
            position: position of the model.
            data: data name mappings.
            latent_size: latent size. Defaults to 2.
        """
        super().__init__(name=name, data=data)
        self.position = position
        self.input_key = f"{name}_{data['input']}"
        self.target_key = f"{name}_{data['target']}"
        self.latent_size = latent_size

    def decode(self, z: Any, *args, **kwargs) -> Any:
        """Decode a latent space point.

        Args:
            z: latent point.

        Returns:
            decoded sample.
        """
        return z

    def encode(self, x: Any, *args, **kwargs) -> Any:
        """Encode a sample.

        Args:
            x: input sample.

        Returns:
            latent encoding.
        """
        return x

    def inference(self, z: Any, *args, **kwargs) -> Any:
        """Run the model in inference mode.

        Args:
            z: sample.

        Returns:
            generated output.
        """
        return z

    def forward(self, x: Any, *args, **kwargs) -> Any:
        """Forward pass in the model.

        Args:
            x: model input.

        Returns:
            model output.
        """
        return x

    def _run_step(self, x: Any, *args, **kwargs) -> Any:
        """Run a step in the model.

        Args:
            x: model input.

        Returns:
            model step output.
        """
        return x

    def encode_decode(self, x: Any, *args, **kwargs) -> Any:
        """Encode and decode a sample.

        Args:
            x: input sample.

        Returns:
            decoded sample.
        """
        z, x_out = self._run_step(x)
        return z, x_out

    def step(
        self,
        input_data: Any,
        target_data: Any,
        device: str = "cpu",
        current_epoch: int = 0,
        *args,
        **kwargs,
    ) -> Tuple[Any, Any, Any]:
        """Training step for the model.

        Args:
            input_data: input for the step.
            target_data: target for the step.
            device: string representing the device to use. Defaults to "cpu".
            current_epoch: current epoch. Defaults to 0.

        Returns:
            a tuple containing the step output, the loss and the logs for the module.
        """
        z = input_data

        loss = 0
        logs = {"loss": loss}

        return z, loss, logs

    def val_step(
        self,
        input_data: Any,
        target_data: Any,
        device: str = "cpu",
        current_epoch: int = 0,
        *args,
        **kwargs,
    ) -> Any:
        """Validation step for the model.

        Args:
            input_data: input for the step.
            target_data: target for the step.
            device: string representing the device to use. Defaults to "cpu".
            current_epoch: current epoch. Defaults to 0.

        Returns:
            a tuple containing the step output, the loss and the logs for the module.
        """
        z = input_data

        loss = 0
        logs = {"loss": loss}
        return z, loss, logs

    @staticmethod
    def add_model_specific_args(
        parent_parser: ArgumentParser, name: str, *args, **kwargs
    ) -> ArgumentParser:
        """Adding to a parser model specific arguments.

        Args:
            parent_parser: patent parser.
            name: model name.

        Returns:
            update parser.
        """
        parser = ArgumentParser(parents=[parent_parser], add_help=False)

        parser.add_argument(f"--data_path_{name}", type=str)
        parser.add_argument(f"--data_file_{name}", type=str)
        parser.add_argument(f"--position_{name}", type=int, nargs="+")
        parser.add_argument(f"--input_{name}", type=str)
        parser.add_argument(f"--target_{name}", type=str)
        parser.add_argument(f"--checkpoint_path_{name}", type=str)
        parser.add_argument(f"--start_from_checkpoint_{name}", type=str2bool)
        parser.add_argument(f"--checkpoint_model_name_{name}", type=str)
        parser.add_argument(f"--latent_size_{name}", type=int)

        return parser
